"""
Function for getting CropScape data based on an entered lat/lon.

"""

import datetime
import requests

try:
    from pyproj import Transformer
except ImportError:
    from pyproj.transformer import Transformer


def croptype(lat=None, lon=None, year=None):
    """
    Function for working with the CropScape API to get a crop type based on
    the lat,lon, and year entered. The lat/lon is converted to the projection
    used by CropScape before pased to the API. Note, the requests library
    is indicating a bad handshake with the server so 'verify' is currently
    set to False which is unsecure. Use at your own risk until it can be
    resolved. CropScape - Copyright © Center For Spatial Information Science
    and Systems 2009 - 2018

    Parameters
    ----------
    lat : float
        Latitude of point to retrieve.
    lon : float
        Longitude of point to retrieve.
    year : int
        Year to get croptype for.

    Returns
    -------
    category : string
        String of the crop type at that specific lat/lon for the given year.

    Examples
    --------
    To get the crop type, simply do:

    .. code-block :: python

        type = act.discovery.get_cropscape.croptype(36.8172,-97.1709,'2018')

    """
    # Return if lat/lon are not passed in
    if lat is None or lon is None:
        raise RuntimeError('Lat and Lon need to be provided')

    # Set the CropScape Projection
    outproj = (
        'PROJCS["NAD_1983_Albers",'
        'GEOGCS["NAD83",'
        'DATUM["North_American_Datum_1983",'
        'SPHEROID["GRS 1980",6378137,298.257222101,'
        'AUTHORITY["EPSG","7019"]],'
        'TOWGS84[0,0,0,0,0,0,0],'
        'AUTHORITY["EPSG","6269"]],'
        'PRIMEM["Greenwich",0,'
        'AUTHORITY["EPSG","8901"]],'
        'UNIT["degree",0.0174532925199433,'
        'AUTHORITY["EPSG","9108"]],'
        'AUTHORITY["EPSG","4269"]],'
        'PROJECTION["Albers_Conic_Equal_Area"],'
        'PARAMETER["standard_parallel_1",29.5],'
        'PARAMETER["standard_parallel_2",45.5],'
        'PARAMETER["latitude_of_center",23],'
        'PARAMETER["longitude_of_center",-96],'
        'PARAMETER["false_easting",0],'
        'PARAMETER["false_northing",0],'
        'UNIT["meters",1]]'
    )

    # Set the input projection to be lat/lon
    inproj = 'EPSG:4326'

    # Get the x/y coordinates for CropScape
    transformer = Transformer.from_crs(inproj, outproj)
    x, y = transformer.transform(lat, lon)

    # Build URL
    url = 'https://nassgeodata.gmu.edu/axis2/services/CDLService/GetCDLValue?'
    if year is None:
        now = datetime.datetime.now()
        year = now.year - 1

    # Add year, lat, and lon as parameters
    params = {'year': str(year), 'x': str(x), 'y': str(y)}

    # Perform the request.  Note, verify set to False until
    # server SSL errors can be worked out
    try:
        req = requests.get(url, params=params, verify=False, timeout=1)
    except Exception:
        return

    # Return from the webservice is not convertable to json
    # So we need to do some text mining
    text = req.text
    text = text.split(',')
    category = [t for t in text if 'category' in t]
    category = category[0].split(': ')[-1][1:-1]

    return category



"""
Functions specifically for working with QC/DQRs from
the Atmospheric Radiation Measurement Program (ARM).

"""

import datetime as dt
import numpy as np
import requests

from act.config import DEFAULT_DATASTREAM_NAME


def add_dqr_to_qc(
    ds,
    variable=None,
    assessment='incorrect,suspect',
    exclude=None,
    include=None,
    normalize_assessment=True,
    cleanup_qc=True,
):
    """
    Function to query the ARM DQR web service for reports and
    add as a new quality control test to ancillary quality control
    variable. If no anicllary quality control variable exist a new
    one will be created and lined to the data variable through
    ancillary_variables attribure.

    See online documentation from ARM Data
    Quality Office on the use of the DQR web service.

    https://code.arm.gov/docs/dqrws-examples/wikis/home

    Information about the DQR web-service avaible at
    https://adc.arm.gov/dqrws/

    Parameters
    ----------
    ds : xarray.Dataset
        Xarray dataset
    variable : string, or list of str, or None
        Variables to check DQR web service. If set to None will
        attempt to update all variables.
    assessment : string
        assessment type to get DQRs. Current options include
        'missing', 'suspect', 'incorrect' or any combination separated
        by a comma.
    exclude : list of strings
        DQR IDs to exclude from adding into QC
    include : list of strings
        List of DQR IDs to include in flagging of data. Any other DQR IDs
        will be ignored.
    normalize_assessment : boolean
        The DQR assessment term is different than the embedded QC
        term. Embedded QC uses "Bad" and "Indeterminate" while
        DQRs use "Incorrect" and "Suspect". Setting this will ensure
        the same terms are used for both.
    cleanup_qc : boolean
        Call clean.cleanup() method to convert to standardized ancillary
        quality control variables. Has a little bit of overhead so
        if the Dataset has already been cleaned up, no need to run.

    Returns
    -------
    ds : xarray.Dataset
        Xarray dataset containing new quality control variables

    Examples
    --------
        .. code-block:: python

            from act.qc.arm import add_dqr_to_qc
            ds = add_dqr_to_qc(ds, variable=['temp_mean', 'atmos_pressure'])


    """

    # DQR Webservice goes off datastreams, pull from the dataset
    if 'datastream' in ds.attrs:
        datastream = ds.attrs['datastream']
    elif '_datastream' in ds.attrs:
        datastream = ds.attrs['_datastream']
    else:
        raise ValueError('Dataset does not have datastream attribute')

    if datastream == DEFAULT_DATASTREAM_NAME:
        raise ValueError("'datastream' name required for DQR service set to default value "
                         f"{datastream}. Unable to perform DQR service query.")

    # Clean up QC to conform to CF conventions
    if cleanup_qc:
        ds.clean.cleanup()

    # In order to properly flag data, get all variables if None. Exclude QC variables.
    if variable is None:
        variable = list(set(ds.data_vars) - set(ds.clean.matched_qc_variables))

    # Check to ensure variable is list
    if not isinstance(variable, (list, tuple)):
        variable = [variable]

    # Loop through each variable and call web service for that variable
    for var_name in variable:
        # Create URL
        url = 'http://www.archive.arm.gov/dqrws/ARMDQR?datastream='
        url += datastream
        url += '&varname=' + var_name
        url += ''.join(
            [
                '&searchmetric=',
                assessment,
                '&dqrfields=dqrid,starttime,endtime,metric,subject',
            ]
        )

        # Call web service
        req = requests.get(url)

        # Check status values and raise error if not successful
        status = req.status_code
        if status == 400:
            raise ValueError('Check parameters')
        if status == 500:
            raise ValueError('DQR Webservice Temporarily Down')

        # Get data and run through each dqr
        dqrs = req.text.splitlines()
        time = ds['time'].values
        dqr_results = {}
        for line in dqrs:
            line = line.split('|')
            dqr_no = line[0]

            # Exclude DQRs if in list
            if exclude is not None and dqr_no in exclude:
                continue

            # Only include if in include list
            if include is not None and dqr_no not in include:
                continue

            starttime = np.datetime64(dt.datetime.utcfromtimestamp(int(line[1])))
            endtime = np.datetime64(dt.datetime.utcfromtimestamp(int(line[2])))
            ind = np.where((time >= starttime) & (time <= endtime))
            if ind[0].size == 0:
                continue

            if dqr_no in dqr_results.keys():
                dqr_results[dqr_no]['index'] = np.append(dqr_results[dqr_no]['index'], ind)
            else:
                dqr_results[dqr_no] = {
                    'index': ind,
                    'test_assessment': line[3],
                    'test_meaning': ': '.join([dqr_no, line[-1]]),
                }

        for key, value in dqr_results.items():
            try:
                ds.qcfilter.add_test(
                    var_name,
                    index=value['index'],
                    test_meaning=value['test_meaning'],
                    test_assessment=value['test_assessment'],
                )
            except IndexError:
                print(f"Skipping '{var_name}' DQR application because of IndexError")

        if normalize_assessment:
            ds.clean.normalize_assessment(variables=var_name)

    return ds



""" Unit tests for ACT utils module. """

import importlib
import tempfile
from datetime import datetime
from pathlib import Path
import tarfile
from os import chdir, PathLike
import string
import random
import numpy as np
import pandas as pd
import pytest
import pytz
import xarray as xr

import act

spec = importlib.util.find_spec('pyart')
if spec is not None:
    PYART_AVAILABLE = True
else:
    PYART_AVAILABLE = False


def test_dates_between():
    start_date = '20191201'
    end_date = '20201201'
    date_list = act.utils.dates_between(start_date, end_date)
    start_string = datetime.strptime(start_date, '%Y%m%d').strftime('%Y-%m-%d')
    end_string = datetime.strptime(end_date, '%Y%m%d').strftime('%Y-%m-%d')
    answer = np.arange(start_string, end_string, dtype='datetime64[D]')
    answer = np.append(answer, answer[-1] + 1)
    answer = answer.astype('datetime64[s]').astype(int)
    answer = [datetime.utcfromtimestamp(ii) for ii in answer]

    assert date_list == answer


def test_add_in_nan():
    # Make a 1D array of 10 minute data
    time = np.arange('2019-01-01T01:00', '2019-01-01T01:10', dtype='datetime64[m]')
    time = time.astype('datetime64[us]')
    time = np.delete(time, range(3, 8))
    data = np.linspace(0.0, 8.0, time.size)

    time_filled, data_filled = act.utils.add_in_nan(xr.DataArray(time), xr.DataArray(data))
    assert isinstance(time_filled, xr.core.dataarray.DataArray)
    assert isinstance(data_filled, xr.core.dataarray.DataArray)

    time_filled, data_filled = act.utils.add_in_nan(time, data)
    assert isinstance(time_filled, np.ndarray)
    assert isinstance(data_filled, np.ndarray)

    assert time_filled[3] == np.datetime64('2019-01-01T01:05:00')
    assert time_filled[4] == np.datetime64('2019-01-01T01:08:00')
    assert np.isnan(data_filled[3])
    assert data_filled[4] == 6.0

    time_filled, data_filled = act.utils.add_in_nan(time[0], data[0])
    assert time_filled == time[0]
    assert data_filled == data[0]

    # Check for multiple instances of missing data periods
    time = np.arange('2019-01-01T01:00', '2019-01-01T02:00', dtype='datetime64[m]')
    time = np.delete(time, range(3, 8))
    time = np.delete(time, range(33, 36))
    data = np.linspace(0.0, 10.0, time.size)

    time_filled, data_filled = act.utils.add_in_nan(time, data)
    assert time_filled.size == 54
    assert data_filled.size == 54
    index = np.where(time_filled == np.datetime64('2019-01-01T01:37'))[0]
    assert index[0] == 33
    assert np.isclose(data_filled[33], 6.27450)
    index = np.where(time_filled == np.datetime64('2019-01-01T01:38'))[0]
    assert index.size == 0
    index = np.where(time_filled == np.datetime64('2019-01-01T01:39'))[0]
    assert index[0] == 34
    assert np.isnan(data_filled[34])
    index = np.where(time_filled == np.datetime64('2019-01-01T01:40'))[0]
    assert index.size == 0


def test_get_missing_value():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    missing = act.utils.data_utils.get_missing_value(
        ds, 'latent_heat_flux', use_FillValue=True, add_if_missing_in_ds=True
    )
    assert missing == -9999

    ds['latent_heat_flux'].attrs['missing_value'] = -9998
    missing = act.utils.data_utils.get_missing_value(ds, 'latent_heat_flux')
    assert missing == -9998


def test_convert_units():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    data = ds['soil_temp_1'].values
    in_units = ds['soil_temp_1'].attrs['units']
    r_data = act.utils.data_utils.convert_units(data, in_units, 'K')
    assert np.ceil(r_data[0]) == 285

    data = act.utils.data_utils.convert_units(r_data, 'K', 'C')
    assert np.ceil(data[0]) == 12

    try:
        ds.utils.change_units()
    except ValueError as error:
        assert str(error) == "Need to provide 'desired_unit' keyword for .change_units() method"

    desired_unit = 'degF'
    skip_vars = [ii for ii in ds.data_vars if ii.startswith('qc_')]
    ds.utils.change_units(
        variables=None,
        desired_unit=desired_unit,
        skip_variables=skip_vars,
        skip_standard=True,
    )
    units = []
    for var_name in ds.data_vars:
        try:
            units.append(ds[var_name].attrs['units'])
        except KeyError:
            pass
    indices = [i for i, x in enumerate(units) if x == desired_unit]
    assert indices == [0, 2, 4, 6, 8, 32, 34, 36, 38, 40]

    var_name = 'home_signal_15'
    desired_unit = 'V'
    ds.utils.change_units(var_name, desired_unit, skip_variables='lat')
    assert ds[var_name].attrs['units'] == desired_unit

    var_names = ['home_signal_15', 'home_signal_30']
    ds.utils.change_units(var_names, desired_unit)
    for var_name in var_names:
        assert ds[var_name].attrs['units'] == desired_unit

    ds.close()
    del ds

    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_CEIL1)
    var_name = 'range'
    desired_unit = 'km'
    ds = ds.utils.change_units(var_name, desired_unit)
    assert ds[var_name].attrs['units'] == desired_unit
    assert np.isclose(np.sum(ds[var_name].values), 952.56, atol=0.01)

    ds.close()
    del ds


def test_ts_weighted_average():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET_WILDCARD)
    cf_ds = {
        'sgpmetE13.b1': {
            'variable': [
                'tbrg_precip_total',
                'org_precip_rate_mean',
                'pwd_precip_rate_mean_1min',
            ],
            'weight': [0.8, 0.15, 0.05],
            'ds': ds,
        }
    }
    data = act.utils.data_utils.ts_weighted_average(cf_ds)

    np.testing.assert_almost_equal(np.sum(data), 84.9, decimal=1)


def test_accum_precip():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET_WILDCARD)

    ds = act.utils.accumulate_precip(ds, 'tbrg_precip_total')
    dmax = round(np.nanmax(ds['tbrg_precip_total_accumulated']))
    assert dmax == 13.0

    ds = act.utils.accumulate_precip(ds, 'tbrg_precip_total', time_delta=60)
    dmax = round(np.nanmax(ds['tbrg_precip_total_accumulated']))
    assert dmax == 13.0

    ds['tbrg_precip_total'].attrs['units'] = 'mm/hr'
    ds = act.utils.accumulate_precip(ds, 'tbrg_precip_total')
    dmax = np.round(np.nanmax(ds['tbrg_precip_total_accumulated']), 2)
    assert dmax == 0.22


def test_calc_cog_sog():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_NAV)

    ds = act.utils.calc_cog_sog(ds)

    cog = ds['course_over_ground'].values
    sog = ds['speed_over_ground'].values

    np.testing.assert_almost_equal(cog[10], 170.987, decimal=3)
    np.testing.assert_almost_equal(sog[15], 0.448, decimal=3)

    ds = ds.rename({'lat': 'latitude', 'lon': 'longitude'})
    ds = act.utils.calc_cog_sog(ds)
    np.testing.assert_almost_equal(cog[10], 170.987, decimal=3)
    np.testing.assert_almost_equal(sog[15], 0.448, decimal=3)


def test_destination_azimuth_distance():
    lat = 37.1509
    lon = -98.362
    lat2, lon2 = act.utils.destination_azimuth_distance(lat, lon, 180.0, 100)

    np.testing.assert_almost_equal(lat2, 37.150, decimal=3)
    np.testing.assert_almost_equal(lon2, -98.361, decimal=3)


def test_calculate_dqr_times():
    ebbr1_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    ebbr2_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_EBBR2)
    brs_ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_BRS)
    ebbr1_result = act.utils.calculate_dqr_times(ebbr1_ds, variable=['soil_temp_1'], threshold=2)
    ebbr2_result = act.utils.calculate_dqr_times(
        ebbr2_ds, variable=['rh_bottom_fraction'], qc_bit=3, threshold=2
    )
    ebbr3_result = act.utils.calculate_dqr_times(
        ebbr2_ds, variable=['rh_bottom_fraction'], qc_bit=3
    )
    brs_result = act.utils.calculate_dqr_times(
        brs_ds, variable='down_short_hemisp_min', qc_bit=2, threshold=30
    )
    assert ebbr1_result == [('2019-11-25 02:00:00', '2019-11-25 04:30:00')]
    assert ebbr2_result == [('2019-11-30 00:00:00', '2019-11-30 11:00:00')]
    assert brs_result == [('2019-07-05 01:57:00', '2019-07-05 11:07:00')]
    assert ebbr3_result is None
    with tempfile.TemporaryDirectory() as tmpdirname:
        write_file = Path(tmpdirname)
        brs_result = act.utils.calculate_dqr_times(
            brs_ds,
            variable='down_short_hemisp_min',
            qc_bit=2,
            threshold=30,
            txt_path=str(write_file),
        )

    brs_result = act.utils.calculate_dqr_times(
        brs_ds,
        variable='down_short_hemisp_min',
        qc_bit=2,
        threshold=30,
        return_missing=False,
    )
    assert len(brs_result[0]) == 2

    ebbr1_ds.close()
    ebbr2_ds.close()
    brs_ds.close()


def test_decode_present_weather():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)
    ds = act.utils.decode_present_weather(ds, variable='pwd_pw_code_inst')

    data = ds['pwd_pw_code_inst_decoded'].values
    result = 'No significant weather observed'
    assert data[0] == result
    assert data[100] == result
    assert data[600] == result

    np.testing.assert_raises(ValueError, act.utils.inst_utils.decode_present_weather, ds)
    np.testing.assert_raises(
        ValueError,
        act.utils.inst_utils.decode_present_weather,
        ds,
        variable='temp_temp',
    )


def test_datetime64_to_datetime():
    time_datetime = [
        datetime(2019, 1, 1, 1, 0),
        datetime(2019, 1, 1, 1, 1),
        datetime(2019, 1, 1, 1, 2),
        datetime(2019, 1, 1, 1, 3),
        datetime(2019, 1, 1, 1, 4),
    ]

    time_datetime64 = [
        np.datetime64(datetime(2019, 1, 1, 1, 0)),
        np.datetime64(datetime(2019, 1, 1, 1, 1)),
        np.datetime64(datetime(2019, 1, 1, 1, 2)),
        np.datetime64(datetime(2019, 1, 1, 1, 3)),
        np.datetime64(datetime(2019, 1, 1, 1, 4)),
    ]

    time_datetime64_to_datetime = act.utils.datetime_utils.datetime64_to_datetime(time_datetime64)
    assert time_datetime == time_datetime64_to_datetime


@pytest.mark.skipif(not PYART_AVAILABLE, reason="Py-ART is not installed.")
def test_create_pyart_obj():
    try:
        ds = act.io.mpl.read_sigma_mplv5(act.tests.EXAMPLE_SIGMA_MPLV5)
    except Exception:
        return

    radar = act.utils.create_pyart_obj(ds, range_var='range')
    variables = list(radar.fields)
    assert 'nrb_copol' in variables
    assert 'nrb_crosspol' in variables
    assert radar.sweep_start_ray_index['data'][-1] == 67
    assert radar.sweep_end_ray_index['data'][-1] == 101
    assert radar.fixed_angle['data'] == 2.0
    assert radar.scan_type == 'ppi'
    assert radar.sweep_mode['data'] == 'ppi'
    np.testing.assert_allclose(radar.sweep_number['data'][-3:], [1.0, 1.0, 1.0])
    np.testing.assert_allclose(radar.sweep_number['data'][0:3], [0.0, 0.0, 0.0])

    # coordinates
    np.testing.assert_allclose(radar.azimuth['data'][0:5], [-95.0, -92.5, -90.0, -87.5, -85.0])
    np.testing.assert_allclose(radar.elevation['data'][0:5], [2.0, 2.0, 2.0, 2.0, 2.0])
    np.testing.assert_allclose(
        radar.range['data'][0:5],
        [14.98962308, 44.96886923, 74.94811538, 104.92736153, 134.90660768],
    )
    gate_lat = radar.gate_latitude['data'][0, 0:5]
    gate_lon = radar.gate_longitude['data'][0, 0:5]
    gate_alt = radar.gate_altitude['data'][0, 0:5]
    np.testing.assert_allclose(
        gate_lat, [38.95293483, 38.95291135, 38.95288786, 38.95286437, 38.95284089]
    )
    np.testing.assert_allclose(
        gate_lon, [-76.8363515, -76.83669666, -76.83704182, -76.83738699, -76.83773215]
    )
    np.testing.assert_allclose(
        gate_alt, [62.84009906, 63.8864653, 64.93293721, 65.9795148, 67.02619806]
    )
    ds.close()
    del radar


def test_add_solar_variable():
    ds = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_NAV)
    new_ds = act.utils.geo_utils.add_solar_variable(ds)

    assert 'sun_variable' in list(new_ds.keys())
    assert new_ds['sun_variable'].values[10] == 1
    assert np.sum(new_ds['sun_variable'].values) >= 598

    new_ds = act.utils.geo_utils.add_solar_variable(ds, dawn_dusk=True)
    assert 'sun_variable' in list(new_ds.keys())
    assert new_ds['sun_variable'].values[10] == 1
    assert np.sum(new_ds['sun_variable'].values) >= 1234

    ds = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_MET1)
    new_ds = act.utils.geo_utils.add_solar_variable(ds, dawn_dusk=True)
    assert np.sum(new_ds['sun_variable'].values) >= 1046

    ds = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_IRTSST)
    ds = ds.fillna(0)
    new_ds = act.utils.geo_utils.add_solar_variable(ds)
    assert np.sum(new_ds['sun_variable'].values) >= 12

    ds = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_IRTSST)
    ds.drop_vars('lat')
    pytest.raises(ValueError, act.utils.geo_utils.add_solar_variable, ds)

    ds = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_IRTSST)
    ds.drop_vars('lon')
    pytest.raises(ValueError, act.utils.geo_utils.add_solar_variable, ds)
    ds.close()
    new_ds.close()


def test_reduce_time_ranges():
    time = pd.date_range(start='2020-01-01T00:00:00', freq='1min', periods=100)
    time = time.to_list()
    time = time[0:50] + time[60:]
    result = act.utils.datetime_utils.reduce_time_ranges(time)
    assert len(result) == 2
    assert result[1][1].minute == 39

    result = act.utils.datetime_utils.reduce_time_ranges(time, broken_barh=True)
    assert len(result) == 2


def test_planck_converter():
    wnum = 1100
    temp = 300
    radiance = 81.5
    result = act.utils.radiance_utils.planck_converter(wnum=wnum, temperature=temp)
    np.testing.assert_almost_equal(result, radiance, decimal=1)
    result = act.utils.radiance_utils.planck_converter(wnum=wnum, radiance=radiance)
    assert np.ceil(result) == temp
    np.testing.assert_raises(ValueError, act.utils.radiance_utils.planck_converter)


def test_solar_azimuth_elevation():

    ds = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_NAV)

    elevation, azimuth, distance = act.utils.geo_utils.get_solar_azimuth_elevation(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        time=ds['time'].values,
        library='skyfield',
        temperature_C='standard',
        pressure_mbar='standard',
    )
    assert np.isclose(np.nanmean(elevation), 10.5648, atol=0.001)
    assert np.isclose(np.nanmean(azimuth), 232.0655, atol=0.001)
    assert np.isclose(np.nanmean(distance), 0.985, atol=0.001)


def test_get_sunrise_sunset_noon():

    ds = act.io.armfiles.read_netcdf(act.tests.EXAMPLE_NAV)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=ds['time'].values[0],
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=ds['time'].values[0],
        library='skyfield',
        timezone=True,
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32, tzinfo=pytz.UTC)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4, tzinfo=pytz.UTC)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10, tzinfo=pytz.UTC)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date='20180201',
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=['20180201'],
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=datetime(2018, 2, 1),
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=datetime(2018, 2, 1, tzinfo=pytz.UTC),
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=ds['lat'].values[0],
        longitude=ds['lon'].values[0],
        date=[datetime(2018, 2, 1)],
        library='skyfield',
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 1, 31, 22, 36, 32)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 2, 1, 17, 24, 4)
    assert noon[0].replace(microsecond=0) == datetime(2018, 2, 1, 8, 2, 10)

    sunrise, sunset, noon = act.utils.geo_utils.get_sunrise_sunset_noon(
        latitude=85.0, longitude=-140.0, date=[datetime(2018, 6, 1)], library='skyfield'
    )
    assert sunrise[0].replace(microsecond=0) == datetime(2018, 3, 30, 10, 48, 48)
    assert sunset[0].replace(microsecond=0) == datetime(2018, 9, 12, 8, 50, 14)
    assert noon[0].replace(microsecond=0) == datetime(2018, 6, 1, 21, 17, 52)


def test_is_sun_visible():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_EBBR1)
    result = act.utils.geo_utils.is_sun_visible(
        latitude=ds['lat'].values,
        longitude=ds['lon'].values,
        date_time=ds['time'].values,
    )
    assert len(result) == 48
    assert sum(result) == 20

    result = act.utils.geo_utils.is_sun_visible(
        latitude=ds['lat'].values,
        longitude=ds['lon'].values,
        date_time=ds['time'].values[0],
    )
    assert result == [False]

    result = act.utils.geo_utils.is_sun_visible(
        latitude=ds['lat'].values,
        longitude=ds['lon'].values,
        date_time=[datetime(2019, 11, 25, 13, 30, 00)],
    )
    assert result == [True]

    result = act.utils.geo_utils.is_sun_visible(
        latitude=ds['lat'].values,
        longitude=ds['lon'].values,
        date_time=datetime(2019, 11, 25, 13, 30, 00),
    )
    assert result == [True]


def test_convert_to_potential_temp():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    temp_var_name = 'temp_mean'
    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.convert_to_potential_temp(
        ds, temp_var_name, press_var_name=press_var_name
    )
    assert np.isclose(np.nansum(temp), -4240.092, rtol=0.001, atol=0.001)
    temp = act.utils.data_utils.convert_to_potential_temp(
        temperature=ds[temp_var_name].values,
        pressure=ds[press_var_name].values,
        temp_var_units=ds[temp_var_name].attrs['units'],
        press_var_units=ds[press_var_name].attrs['units'],
    )
    assert np.isclose(np.nansum(temp), -4240.092, rtol=0.001, atol=0.0011)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.convert_to_potential_temp(
            temperature=ds[temp_var_name].values,
            pressure=ds[press_var_name].values,
            temp_var_units=ds[temp_var_name].attrs['units'],
        )

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.convert_to_potential_temp(
            temperature=ds[temp_var_name].values,
            pressure=ds[press_var_name].values,
            press_var_units=ds[press_var_name].attrs['units'],
        )


def test_height_adjusted_temperature():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    temp_var_name = 'temp_mean'
    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.height_adjusted_temperature(
        ds,
        temp_var_name,
        height_difference=100,
        height_units='m',
        press_var_name=press_var_name,
    )
    assert np.isclose(np.nansum(temp), -6834.291, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        ds, temp_var_name=temp_var_name, height_difference=-900, height_units='feet'
    )
    assert np.isclose(np.nansum(temp), -1904.7257, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        ds,
        temp_var_name,
        height_difference=-200,
        height_units='m',
        press_var_name=press_var_name,
        pressure=102.325,
        press_var_units='kPa',
    )
    assert np.isclose(np.nansum(temp), -2871.5435, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_temperature(
        height_difference=25.2,
        height_units='m',
        temperature=ds[temp_var_name].values,
        temp_var_units=ds[temp_var_name].attrs['units'],
        pressure=ds[press_var_name].values,
        press_var_units=ds[press_var_name].attrs['units'],
    )
    assert np.isclose(np.nansum(temp), -5847.511, rtol=0.001, atol=0.001)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.height_adjusted_temperature(
            height_difference=25.2,
            height_units='m',
            temperature=ds[temp_var_name].values,
            temp_var_units=None,
            pressure=ds[press_var_name].values,
            press_var_units=ds[press_var_name].attrs['units'],
        )


def test_height_adjusted_pressure():
    ds = act.io.armfiles.read_netcdf(act.tests.sample_files.EXAMPLE_MET1)

    press_var_name = 'atmos_pressure'
    temp = act.utils.data_utils.height_adjusted_pressure(
        ds=ds, press_var_name=press_var_name, height_difference=20, height_units='m'
    )
    assert np.isclose(np.nansum(temp), 142020.83, rtol=0.001, atol=0.001)

    temp = act.utils.data_utils.height_adjusted_pressure(
        height_difference=-100,
        height_units='ft',
        pressure=ds[press_var_name].values,
        press_var_units=ds[press_var_name].attrs['units'],
    )
    assert np.isclose(np.nansum(temp), 142877.69, rtol=0.001, atol=0.001)

    with np.testing.assert_raises(ValueError):
        temp = act.utils.data_utils.height_adjusted_pressure(
            height_difference=-100,
            height_units='ft',
            pressure=ds[press_var_name].values,
            press_var_units=None,
        )


def test_date_parser():
    datestring = '20111001'
    output_format = '%Y/%m/%d'

    test_string = act.utils.date_parser(datestring, output_format, return_datetime=False)
    assert test_string == '2011/10/01'

    test_datetime = act.utils.date_parser(datestring, output_format, return_datetime=True)
    assert test_datetime == datetime(2011, 10, 1)


def test_date_parser_minute_second():
    date_string = '2020-01-01T12:00:00'
    parsed_date = act.utils.date_parser(date_string, return_datetime=True)
    assert parsed_date == datetime(2020, 1, 1, 12, 0, 0)

    output_format = parsed_date.strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z'
    assert output_format == '2020-01-01T12:00:00.000Z'


def test_adjust_timestamp():
    file = act.tests.sample_files.EXAMPLE_EBBR1
    ds = act.io.armfiles.read_netcdf(file)
    ds = act.utils.datetime_utils.adjust_timestamp(ds)
    assert ds['time'].values[0] == np.datetime64('2019-11-24T23:30:00.000000000')

    ds = act.utils.datetime_utils.adjust_timestamp(ds, offset=-60 * 60)
    assert ds['time'].values[0] == np.datetime64('2019-11-24T22:30:00.000000000')



"""
Stores the class for TimeSeriesDisplay.

"""

import datetime as dt
import textwrap
import warnings
from copy import deepcopy
from re import search, search as re_search

import matplotlib as mpl
import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import colors as mplcolors
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.interpolate import NearestNDInterpolator

from ..qc.qcfilter import parse_bit
from ..utils import data_utils, datetime_utils as dt_utils
from ..utils.datetime_utils import determine_time_delta, reduce_time_ranges
from ..utils.geo_utils import get_sunrise_sunset_noon
from . import common
from .plot import Display


class TimeSeriesDisplay(Display):
    """
    This subclass contains routines that are specific to plotting
    time series plots from data. It is inherited from Display and therefore
    contains all of Display's attributes and methods.

    Examples
    --------

    To create a TimeSeriesDisplay with 3 rows, simply do:

    .. code-block:: python

        ds = act.read_netcdf(the_file)
        disp = act.plotting.TimeSeriesDisplay(ds, subplot_shape=(3,), figsize=(15, 5))

    The TimeSeriesDisplay constructor takes in the same keyword arguments as
    plt.subplots. For more information on the plt.subplots keyword arguments,
    see the `matplotlib documentation
    <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.subplots.html>`_.
    If no subplot_shape is provided, then no figure or axis will be created
    until add_subplots or plots is called.

    """

    def __init__(self, ds, subplot_shape=(1,), ds_name=None, **kwargs):
        super().__init__(ds, subplot_shape, ds_name, **kwargs)

    def day_night_background(self, dsname=None, subplot_index=(0,)):
        """
        Colorcodes the background according to sunrise/sunset.

        Parameters
        ----------
        dsname : None or str
            If there is more than one datastream in the display object the
            name of the datastream needs to be specified. If set to None and
            there is only one datastream then ACT will use the sole datastream
            in the object.
        subplot_index : 1 or 2D tuple, list, or array
            The index to the subplot to place the day and night background in.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream to derive the '
                + 'information needed for the day and night '
                + 'background when 2 or more datasets are in '
                + 'the display object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Get File Dates
        try:
            file_dates = self._ds[dsname].attrs['_file_dates']
        except KeyError:
            file_dates = []
        if len(file_dates) == 0:
            sdate = dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0])
            edate = dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[-1])
            file_dates = [sdate, edate]

        all_dates = dt_utils.dates_between(file_dates[0], file_dates[-1])

        if self.axes is None:
            raise RuntimeError('day_night_background requires the plot to ' 'be displayed.')

        ax = self.axes[subplot_index]

        # Find variable names for latitude and longitude
        variables = list(self._ds[dsname].data_vars)
        lat_name = [var for var in ['lat', 'latitude'] if var in variables]
        lon_name = [var for var in ['lon', 'longitude'] if var in variables]
        if len(lat_name) == 0:
            lat_name = None
        else:
            lat_name = lat_name[0]

        if len(lon_name) == 0:
            lon_name = None
        else:
            lon_name = lon_name[0]

        # Variable name does not match, look for standard_name declaration
        if lat_name is None or lon_name is None:
            for var in variables:
                try:
                    if self._ds[dsname][var].attrs['standard_name'] == 'latitude':
                        lat_name = var
                except KeyError:
                    pass

                try:
                    if self._ds[dsname][var].attrs['standard_name'] == 'longitude':
                        lon_name = var
                except KeyError:
                    pass

                if lat_name is not None and lon_name is not None:
                    break

        if lat_name is None or lon_name is None:
            return

        # Extract latitude and longitude scalar from variable. If variable is a vector look
        # for first non-Nan value.
        lat_lon_list = [np.nan, np.nan]
        for ii, var_name in enumerate([lat_name, lon_name]):
            try:
                values = self._ds[dsname][var_name].values
                if values.size == 1:
                    lat_lon_list[ii] = float(values)
                else:
                    # Look for non-NaN values to use for latitude locaiton. If not found use first value.
                    index = np.where(np.isfinite(values))[0]
                    if index.size == 0:
                        lat_lon_list[ii] = float(values[0])
                    else:
                        lat_lon_list[ii] = float(values[index[0]])
            except AttributeError:
                pass

        for value, name in zip(lat_lon_list, ['Latitude', 'Longitude']):
            if not np.isfinite(value):
                warnings.warn(f"{name} value in dataset equal to '{value}' is not finite. ", RuntimeWarning)
                return

        lat = lat_lon_list[0]
        lon = lat_lon_list[1]

        lat_range = [-90, 90]
        if not (lat_range[0] <= lat <= lat_range[1]):
            warnings.warn(
                f"Latitude value in dataset of '{lat}' not within acceptable "
                f'range of {lat_range[0]} <= latitude <= {lat_range[1]}. ',
                RuntimeWarning,
            )
            return

        lon_range = [-180, 180]
        if not (lon_range[0] <= lon <= lon_range[1]):
            warnings.warn(
                f"Longitude value in dataset of '{lon}' not within acceptable "
                f'range of {lon_range[0]} <= longitude <= {lon_range[1]}. ',
                RuntimeWarning,
            )
            return

        # Initialize the plot to a gray background for total darkness
        rect = ax.patch
        rect.set_facecolor('0.85')

        # Get date ranges to plot
        plot_dates = []
        for f in all_dates:
            for ii in [-1, 0, 1]:
                plot_dates.append(f + dt.timedelta(days=ii))

        # Get sunrise, sunset and noon times
        sunrise, sunset, noon = get_sunrise_sunset_noon(lat, lon, plot_dates)

        # Plot daylight
        for ii in range(0, len(sunrise)):
            ax.axvspan(sunrise[ii], sunset[ii], facecolor='#FFFFCC', zorder=0)

        # Plot noon line
        for ii in noon:
            ax.axvline(x=ii, linestyle='--', color='y', zorder=1)

    def set_xrng(self, xrng, subplot_index=(0,)):
        """
        Sets the x range of the plot.

        Parameters
        ----------
        xrng : 2 number array
            The x limits of the plot.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.

        """
        if self.axes is None:
            raise RuntimeError('set_xrng requires the plot to be displayed.')

        # If the xlim is set to the same value for range it will throw a warning
        # This is to catch that and expand the range so we avoid the warning.
        if xrng[0] == xrng[1]:
            if isinstance(xrng[0], np.ndarray) and np.issubdtype(xrng[0].dtype, np.datetime64):
                print(f'\nAttempting to set xlim range to single value {xrng[0]}. '
                      'Expanding range by 2 seconds.\n')
                xrng[0] = xrng[0] - np.timedelta64(1, 's')
                xrng[1] = xrng[1] + np.timedelta64(1, 's')
            elif isinstance(xrng[0], dt.datetime):
                print(f'\nAttempting to set xlim range to single value {xrng[0]}. '
                      'Expanding range by 2 seconds.\n')
                xrng[0] = xrng[0] - dt.timedelta(seconds=1)
                xrng[1] = xrng[1] + dt.timedelta(seconds=1)

        self.axes[subplot_index].set_xlim(xrng)

        # Make sure that the xrng value is a numpy array not pandas
        if isinstance(xrng[0], pd.Timestamp):
            xrng = [x.to_numpy() for x in xrng if isinstance(x, pd.Timestamp)]

        # Make sure that the xrng value is a numpy array not datetime.datetime
        if isinstance(xrng[0], dt.datetime):
            xrng = [np.datetime64(x) for x in xrng if isinstance(x, dt.datetime)]

        if len(subplot_index) < 2:
            self.xrng[subplot_index, 0] = xrng[0].astype('datetime64[D]').astype(float)
            self.xrng[subplot_index, 1] = xrng[1].astype('datetime64[D]').astype(float)
        else:
            self.xrng[subplot_index][0] = xrng[0].astype('datetime64[D]').astype(float)
            self.xrng[subplot_index][1] = xrng[1].astype('datetime64[D]').astype(float)

    def set_yrng(self, yrng, subplot_index=(0,)):
        """
        Sets the y range of the plot.

        Parameters
        ----------
        yrng : 2 number array
            The y limits of the plot.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.

        """
        if self.axes is None:
            raise RuntimeError('set_yrng requires the plot to be displayed.')

        if not hasattr(self, 'yrng') and len(self.axes.shape) == 2:
            self.yrng = np.zeros((self.axes.shape[0], self.axes.shape[1], 2))
        elif not hasattr(self, 'yrng') and len(self.axes.shape) == 1:
            self.yrng = np.zeros((self.axes.shape[0], 2))

        if yrng[0] == yrng[1]:
            yrng[1] = yrng[1] + 1

        self.axes[subplot_index].set_ylim(yrng)
        try:
            self.yrng[subplot_index, :] = yrng
        except IndexError:
            self.yrng[subplot_index] = yrng

    def plot(
        self,
        field,
        dsname=None,
        subplot_index=(0,),
        cmap=None,
        set_title=None,
        add_nan=False,
        day_night_background=False,
        invert_y_axis=False,
        abs_limits=(None, None),
        time_rng=None,
        y_rng=None,
        use_var_for_y=None,
        set_shading='auto',
        assessment_overplot=False,
        overplot_marker='.',
        overplot_behind=False,
        overplot_markersize=6,
        assessment_overplot_category={
            'Incorrect': ['Bad', 'Incorrect'],
            'Suspect': ['Indeterminate', 'Suspect'],
        },
        assessment_overplot_category_color={'Incorrect': 'red', 'Suspect': 'orange'},
        force_line_plot=False,
        labels=False,
        cbar_label=None,
        cbar_h_adjust=None,
        secondary_y=False,
        y_axis_flag_meanings=False,
        colorbar_labels=None,
        cb_friendly=False,
        match_line_label_color=False,
        **kwargs,
    ):
        """
        Makes a timeseries plot. If subplots have not been added yet, an axis
        will be created assuming that there is only going to be one plot.

        If plotting a high data volume 2D dataset, it may take some time to plot.
        In order to speed up your plot creation, please resample your data to a
        lower resolution dataset.

        Parameters
        ----------
        field : str
            The name of the field to plot.
        dsname : None or str
            If there is more than one datastream in the display object the
            name of the datastream needs to be specified. If set to None and
            there is only one datastream ACT will use the sole datastream
            in the object.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.
        cmap : matplotlib colormap
            The colormap to use.
        set_title : str
            The title for the plot.
        add_nan : bool
            Set to True to fill in data gaps with NaNs.
        day_night_background : bool
            Set to True to fill in a color coded background.
            according to the time of day.
        abs_limits : tuple or list
            Sets the bounds on plot limits even if data values exceed
            those limits. Set to (ymin,ymax). Use None if only setting
            minimum or maximum limit, i.e. (22., None).
        time_rng : tuple or list
            List or tuple with (min, max) values to set the x-axis range
            limits.
        y_rng : tuple or list
            List or tuple with (min, max) values to set the y-axis range
        use_var_for_y : str
            Set this to the name of a data variable in the Dataset to use as
            the y-axis variable instead of the default dimension. Useful for
            instances where data has an index-based dimension instead of a
            height-based dimension. If shapes of arrays do not match it will
            automatically revert back to the original ydata.
        set_shading : string
            Option to to set the matplotlib.pcolormesh shading parameter.
            Default to 'auto'
        assessment_overplot : boolean
            Option to overplot quality control colored symbols over plotted
            data using flag_assessment categories.
        overplot_marker : str
            Marker to use for overplot symbol.
        overplot_behind : bool
            Place the overplot marker behind the data point.
        overplot_markersize : float or int
            Size of overplot marker. If overplot_behind or force_line_plot
            are set the marker size will be double overplot_markersize so
            the color is visible.
        assessment_overplot_category : dict
            Lookup to categorize assessments into groups. This allows using
            multiple terms for the same quality control level of failure.
            Also allows adding more to the defaults.
        assessment_overplot_category_color : dict
            Lookup to match overplot category color to assessment grouping.
        force_line_plot : boolean
            Option to plot 2D data as 1D line plots.
        labels : boolean or list
            Option to overwrite the legend labels. Must have same dimensions as
            number of lines plotted.
        cbar_label : str
            Option to overwrite default colorbar label.
        cbar_h_adjust : float
            Option to adjust location of colorbar horizontally. Positive values
            move to right negative values move to left.
        secondary_y : boolean
            Option to plot on secondary y axis.
        y_axis_flag_meanings : boolean or int
            When set to True and plotting state variable with flag_values and
            flag_meanings attributes will replace y axis numerical values
            with flag_meanings value. Set to a positive number larger than 1
            to indicate maximum word length to use. If text is longer that the
            value and has space characters will split text over multiple lines.
        colorbar_labels : dict
            A dictionary containing values for plotting a 2D array of state variables.
            The dictionary uses data values as keys and a dictionary containing keys
            'text' and 'color' for each data value to plot.

            Example:
                {0: {'text': 'Clear sky', 'color': 'white'},
                1: {'text': 'Liquid', 'color': 'green'},
                2: {'text': 'Ice', 'color': 'blue'},
                3: {'text': 'Mixed phase', 'color': 'purple'}}
        cb_friendly : boolean
            Set to true if you want to use the integrated colorblind friendly
            colors for green/red based on the Homeyer colormap.
        match_line_label_color : boolean
            Will set the y label to match the line color in the plot. This
            will only work if the time series plot is a line plot.
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.plot` (1D timeseries) or
            :func:`plt.pcolormesh` (2D timeseries).

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        if y_axis_flag_meanings:
            kwargs['linestyle'] = ''

        if cb_friendly:
            cmap = 'act_HomeyerRainbow'
            assessment_overplot_category_color['Bad'] = (0.9285714285714286, 0.7130901016453677, 0.7130901016453677)
            assessment_overplot_category_color['Incorrect'] = (0.9285714285714286, 0.7130901016453677, 0.7130901016453677)
            assessment_overplot_category_color['Not Failing'] = (0.0, 0.4240129715562796, 0.4240129715562796),
            assessment_overplot_category_color['Acceptable'] = (0.0, 0.4240129715562796, 0.4240129715562796),

        # Get data and dimensions
        data = self._ds[dsname][field]
        dim = list(self._ds[dsname][field].dims)
        xdata = self._ds[dsname][dim[0]]

        if 'units' in data.attrs:
            ytitle = ''.join(['(', data.attrs['units'], ')'])
        else:
            ytitle = field

        if cbar_label is None:
            cbar_default = ytitle
        if len(dim) > 1:
            if use_var_for_y is None:
                ydata = self._ds[dsname][dim[1]]
            else:
                ydata = self._ds[dsname][use_var_for_y]
                ydata_dim1 = self._ds[dsname][dim[1]]
                if np.shape(ydata) != np.shape(ydata_dim1):
                    ydata = ydata_dim1
            units = ytitle
            if 'units' in ydata.attrs.keys():
                units = ydata.attrs['units']
                ytitle = ''.join(['(', units, ')'])
            else:
                units = ''
                ytitle = dim[1]

            # Create labels if 2d as 1d
            if force_line_plot is True:
                if labels is True:
                    labels = [' '.join([str(d), units]) for d in ydata.values]
                ytitle = f"({data.attrs['units']})"
                ydata = None
        else:
            ydata = None

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        # Set up secondary y axis if requested
        if secondary_y is False:
            ax = self.axes[subplot_index]
        else:
            ax = self.axes[subplot_index].twinx()

        if colorbar_labels is not None:
            flag_values = list(colorbar_labels.keys())
            flag_meanings = [value['text'] for key, value in colorbar_labels.items()]
            cbar_colors = [value['color'] for key, value in colorbar_labels.items()]
            cmap = mpl.colors.ListedColormap(cbar_colors)
            for ii, flag_meaning in enumerate(flag_meanings):
                if len(flag_meaning) > 20:
                    flag_meaning = textwrap.fill(flag_meaning, width=20)
                    flag_meanings[ii] = flag_meaning
        else:
            flag_values = None
            flag_meanings = None
            cbar_colors = None

        if ydata is None:
            # Add in nans to ensure the data does not connect the line.
            if add_nan is True:
                xdata, data = data_utils.add_in_nan(xdata, data)

            if day_night_background is True:
                self.day_night_background(subplot_index=subplot_index, dsname=dsname)

            # If limiting data being plotted use masked arrays
            # Need to do it this way because of autoscale() method
            if abs_limits[0] is not None and abs_limits[1] is not None:
                data = np.ma.masked_outside(data, abs_limits[0], abs_limits[1])
            elif abs_limits[0] is not None and abs_limits[1] is None:
                data = np.ma.masked_less_equal(data, abs_limits[0])
            elif abs_limits[0] is None and abs_limits[1] is not None:
                data = np.ma.masked_greater_equal(data, abs_limits[1])

            # Plot the data
            if 'marker' not in kwargs.keys():
                kwargs['marker'] = '.'

            lines = ax.plot(xdata, data, **kwargs)

            # Check if we need to call legend method after plotting. This is only
            # called when no assessment overplot is called.
            add_legend = False
            if 'label' in kwargs.keys():
                add_legend = True

            # Overplot failing data if requested
            if assessment_overplot:
                # If we are doing forced line plot from 2D data need to manage
                # legend lables. Will make arrays to hold labels of QC failing
                # because not set when labels not set.
                if not isinstance(labels, list) and add_legend is False:
                    labels = []
                    lines = []

                # For forced line plot need to plot QC behind point instead of
                # on top of point.
                zorder = None
                if force_line_plot or overplot_behind:
                    zorder = 0
                    overplot_markersize *= 2.0

                for assessment, categories in assessment_overplot_category.items():
                    flag_data = self._ds[dsname].qcfilter.get_masked_data(
                        field, rm_assessments=categories, return_inverse=True
                    )
                    if np.invert(flag_data.mask).any() and np.isfinite(flag_data).any():
                        try:
                            flag_data.mask = np.logical_or(data.mask, flag_data.mask)
                        except AttributeError:
                            pass
                        qc_ax = ax.plot(
                            xdata,
                            flag_data,
                            marker=overplot_marker,
                            linestyle='',
                            markersize=overplot_markersize,
                            color=assessment_overplot_category_color[assessment],
                            label=assessment,
                            zorder=zorder,
                        )
                        # If labels keyword is set need to add labels for calling legend
                        if isinstance(labels, list):
                            # If plotting forced_line_plot need to subset the Line2D object
                            # so we don't have more than one added to legend.
                            if len(qc_ax) > 1:
                                lines.extend(qc_ax[:1])
                            else:
                                lines.extend(qc_ax)
                            labels.append(assessment)
                        add_legend = True

            # Add legend if labels are available
            if isinstance(labels, list):
                ax.legend(lines, labels)
            elif add_legend:
                ax.legend()

            # Change y axis to text from flag_meanings if requested.
            if y_axis_flag_meanings:
                flag_meanings = self._ds[dsname][field].attrs['flag_meanings']
                flag_values = self._ds[dsname][field].attrs['flag_values']
                # If keyword is larger than 1 assume this is the maximum character length
                # desired and insert returns to wrap text.
                if y_axis_flag_meanings > 1:
                    for ii, flag_meaning in enumerate(flag_meanings):
                        if len(flag_meaning) > y_axis_flag_meanings:
                            flag_meaning = textwrap.fill(flag_meaning, width=y_axis_flag_meanings)
                            flag_meanings[ii] = flag_meaning

                ax.set_yticks(flag_values)
                ax.set_yticklabels(flag_meanings)

        else:
            # Add in nans to ensure the data are not streaking
            if add_nan is True:
                xdata, data = data_utils.add_in_nan(xdata, data)

            # Sets shading parameter to auto. Matplotlib will check deminsions.
            # If X,Y and C are same deminsions shading is set to nearest.
            # If X and Y deminsions are 1 greater than C shading is set to flat.
            if 'edgecolors' not in kwargs.keys():
                kwargs['edgecolors'] = 'face'
            mesh = ax.pcolormesh(
                np.asarray(xdata),
                ydata,
                data.transpose(),
                shading=set_shading,
                cmap=cmap,
                **kwargs,
            )

        # Set Title
        if set_title is None:
            if isinstance(self._ds[dsname].time.values[0], np.datetime64):
                set_title = ' '.join(
                    [
                        dsname,
                        field,
                        'on',
                        dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                    ]
                )
            else:
                date_result = search(
                    r'\d{4}-\d{1,2}-\d{1,2}', self._ds[dsname].time.attrs['units']
                )
                if date_result is not None:
                    set_title = ' '.join([dsname, field, 'on', date_result.group(0)])
                else:
                    set_title = ' '.join([dsname, field])

        if secondary_y is False:
            ax.set_title(set_title)

        # Set YTitle
        if not y_axis_flag_meanings:
            if match_line_label_color and len(ax.get_lines()) > 0:
                ax.set_ylabel(ytitle, color=ax.get_lines()[0].get_color())
            else:
                ax.set_ylabel(ytitle)

        # Set X Limit - We want the same time axes for all subplots
        if not hasattr(self, 'time_rng'):
            if time_rng is not None:
                self.time_rng = list(time_rng)
            else:
                self.time_rng = [xdata.min().values, xdata.max().values]

        self.set_xrng(self.time_rng, subplot_index)

        # Set Y Limit
        if y_rng is not None:
            self.set_yrng(y_rng)

        if hasattr(self, 'yrng'):
            # Make sure that the yrng is not just the default
            if ydata is None:
                if abs_limits[0] is not None or abs_limits[1] is not None:
                    our_data = data
                else:
                    our_data = data.values
            else:
                our_data = ydata

            finite = np.isfinite(our_data)
            if finite.any():
                our_data = our_data[finite]
                if invert_y_axis is False:
                    yrng = [np.min(our_data), np.max(our_data)]
                else:
                    yrng = [np.max(our_data), np.min(our_data)]
            else:
                yrng = [0, 1]

            # Check if current range is outside of new range an only set
            # values that work for all data plotted.
            if isinstance(yrng[0], np.datetime64):
                yrng = mdates.datestr2num([str(yrng[0]), str(yrng[1])])

            current_yrng = ax.get_ylim()
            if invert_y_axis is False:
                if yrng[0] > current_yrng[0]:
                    yrng[0] = current_yrng[0]
                if yrng[1] < current_yrng[1]:
                    yrng[1] = current_yrng[1]
            else:
                if yrng[0] < current_yrng[0]:
                    yrng[0] = current_yrng[0]
                if yrng[1] > current_yrng[1]:
                    yrng[1] = current_yrng[1]

            # Set y range the normal way if not secondary y
            # If secondary, just use set_ylim
            if secondary_y is False:
                self.set_yrng(yrng, subplot_index)
            else:
                ax.set_ylim(yrng)

        # Set X Format
        if len(subplot_index) == 1:
            days = self.xrng[subplot_index, 1] - self.xrng[subplot_index, 0]
        else:
            days = self.xrng[subplot_index][1] - self.xrng[subplot_index][0]

        myFmt = common.get_date_format(days)
        ax.xaxis.set_major_formatter(myFmt)

        # Set X format - We want the same time axes for all subplots
        if not hasattr(self, 'time_fmt'):
            self.time_fmt = myFmt

        # Put on an xlabel, but only if we are making the bottom-most plot
        if subplot_index[0] == self.axes.shape[0] - 1:
            ax.set_xlabel('Time [UTC]')

        if ydata is not None:
            if cbar_label is None:
                cbar_title = cbar_default
            else:
                cbar_title = ''.join(['(', cbar_label, ')'])

            if colorbar_labels is not None:
                cbar_title = None
                cbar = self.add_colorbar(
                    mesh,
                    title=cbar_title,
                    subplot_index=subplot_index,
                    values=flag_values,
                    pad=cbar_h_adjust,
                )
                cbar.set_ticks(flag_values)
                cbar.set_ticklabels(flag_meanings)
                cbar.ax.tick_params(labelsize=10)

            else:
                self.add_colorbar(
                    mesh, title=cbar_title, subplot_index=subplot_index, pad=cbar_h_adjust
                )

        return ax

    def plot_barbs_from_spd_dir(
        self, speed_field, direction_field, pres_field=None, dsname=None, **kwargs
    ):
        """
        This procedure will make a wind barb plot timeseries.
        If a pressure field is given and the wind fields are 1D, which, for
        example, would occur if one wants to plot a timeseries of
        rawinsonde data, then a time-height cross section of
        winds will be made.

        Note: This procedure calls plot_barbs_from_u_v and will take in the
        same keyword arguments as that procedure.

        Parameters
        ----------
        speed_field : str
            The name of the field specifying the wind speed in m/s.
        direction_field : str
            The name of the field specifying the wind direction in degrees.
            0 degrees is defined to be north and increases clockwise like
            what is used in standard meteorological notation.
        pres_field : str
            The name of the field specifying pressure or height. If using
            height coordinates, then we recommend setting invert_y_axis
            to False.
        dsname : str
            The name of the datastream to plot. Setting to None will make
            ACT attempt to autodetect this.
        kwargs : dict
            Any additional keyword arguments will be passed into
            :func:`act.plotting.TimeSeriesDisplay.plot_barbs_from_u_and_v`.

        Returns
        -------
        the_ax : matplotlib axis handle
            The handle to the axis where the plot was made on.

        Examples
        --------
        ..code-block :: python

            sonde_ds = act.io.armfiles.read_netcdf(
                act.tests.sample_files.EXAMPLE_TWP_SONDE_WILDCARD)
            BarbDisplay = act.plotting.TimeSeriesDisplay(
                {'sonde_darwin': sonde_ds}, figsize=(10,5))
            BarbDisplay.plot_barbs_from_spd_dir('deg', 'wspd', 'pres',
                                                num_barbs_x=20)

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Make temporary field called tempu, tempv
        spd = self._ds[dsname][speed_field]
        dir = self._ds[dsname][direction_field]
        tempu = -np.sin(np.deg2rad(dir)) * spd
        tempv = -np.cos(np.deg2rad(dir)) * spd
        self._ds[dsname]['temp_u'] = deepcopy(self._ds[dsname][speed_field])
        self._ds[dsname]['temp_v'] = deepcopy(self._ds[dsname][speed_field])
        self._ds[dsname]['temp_u'].values = tempu
        self._ds[dsname]['temp_v'].values = tempv
        the_ax = self.plot_barbs_from_u_v('temp_u', 'temp_v', pres_field, dsname, **kwargs)
        del self._ds[dsname]['temp_u'], self._ds[dsname]['temp_v']
        return the_ax

    def plot_barbs_from_u_v(
        self,
        u_field,
        v_field,
        pres_field=None,
        dsname=None,
        subplot_index=(0,),
        set_title=None,
        day_night_background=False,
        invert_y_axis=True,
        num_barbs_x=20,
        num_barbs_y=20,
        use_var_for_y=None,
        **kwargs,
    ):
        """
        This function will plot a wind barb timeseries from u and v wind
        data. If pres_field is given, a time-height series will be plotted
        from 1-D wind data.

        Parameters
        ----------
        u_field : str
            The name of the field containing the U component of the wind.
        v_field : str
            The name of the field containing the V component of the wind.
        pres_field : str or None
            The name of the field containing the pressure or height. Set
            to None to not use this.
        dsname : str or None
            The name of the datastream to plot. Setting to None will make
            ACT automatically try to determine this.
        subplot_index : 2-tuple
            The index of the subplot to make the plot on.
        set_title : str or None
            The title of the plot.
        day_night_background : bool
            Set to True to plot a day/night background.
        invert_y_axis : bool
            Set to True to invert the y axis (i.e. for plotting pressure as
            the height coordinate).
        num_barbs_x : int
            The number of wind barbs to plot in the x axis.
        num_barbs_y : int
            The number of wind barbs to plot in the y axis.
        cmap : matplotlib.colors.LinearSegmentedColormap
            A color map to use with wind barbs. If this is set the plt.barbs
            routine will be passed the C parameter scaled as sqrt of sum of the
            squares and used with the passed in color map. A colorbar will also
            be added. Setting the limits of the colorbar can be done with 'clim'.
            Setting this changes the wind barbs from black to colors.
        use_var_for_y : str
            Set this to the name of a data variable in the Dataset to use as the
            y-axis variable instead of the default dimension. Useful for instances
            where data has an index-based dimension instead of a height-based
            dimension. If shapes of arrays do not match it will automatically
            revert back to the original ydata.
        **kwargs : keyword arguments
            Additional keyword arguments will be passed into plt.barbs.

        Returns
        -------
        ax : matplotlib axis handle
             The axis handle that contains the reference to the
             constructed plot.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Get data and dimensions
        u = self._ds[dsname][u_field].values
        v = self._ds[dsname][v_field].values
        dim = list(self._ds[dsname][u_field].dims)
        xdata = self._ds[dsname][dim[0]].values
        num_x = xdata.shape[-1]
        barb_step_x = round(num_x / num_barbs_x)
        if barb_step_x == 0:
            barb_step_x = 1

        if len(dim) > 1 and pres_field is None:
            if use_var_for_y is None:
                ydata = self._ds[dsname][dim[1]]
            else:
                ydata = self._ds[dsname][use_var_for_y]
                ydata_dim1 = self._ds[dsname][dim[1]]
                if np.shape(ydata) != np.shape(ydata_dim1):
                    ydata = ydata_dim1
            if 'units' in ydata.attrs:
                units = ydata.attrs['units']
            else:
                units = ''
            ytitle = ''.join(['(', units, ')'])
            num_y = ydata.shape[0]
            barb_step_y = round(num_y / num_barbs_y)
            if barb_step_y == 0:
                barb_step_y = 1

            xdata, ydata = np.meshgrid(xdata, ydata, indexing='ij')
        elif pres_field is not None:
            # What we will do here is do a nearest-neighbor interpolation
            # for each member of the series. Coordinates are time, pressure
            pres = self._ds[dsname][pres_field]
            u_interp = NearestNDInterpolator((xdata, pres.values), u, rescale=True)
            v_interp = NearestNDInterpolator((xdata, pres.values), v, rescale=True)
            barb_step_x = 1
            barb_step_y = 1
            x_times = pd.date_range(xdata.min(), xdata.max(), periods=num_barbs_x)
            if num_barbs_y == 1:
                y_levels = pres.mean()
            else:
                y_levels = np.linspace(np.nanmin(pres), np.nanmax(pres), num_barbs_y)
            xdata, ydata = np.meshgrid(x_times, y_levels, indexing='ij')
            u = u_interp(xdata, ydata)
            v = v_interp(xdata, ydata)
            if 'units' in pres.attrs:
                units = pres.attrs['units']
            else:
                units = ''
            ytitle = ''.join(['(', units, ')'])
        else:
            ydata = None

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        if ydata is None:
            ydata = np.ones(xdata.shape)
            if 'cmap' in kwargs.keys():
                map_color = np.sqrt(np.power(u[::barb_step_x], 2) + np.power(v[::barb_step_x], 2))
                map_color[np.isnan(map_color)] = 0
                ax = self.axes[subplot_index].barbs(
                    xdata[::barb_step_x],
                    ydata[::barb_step_x],
                    u[::barb_step_x],
                    v[::barb_step_x],
                    map_color,
                    **kwargs,
                )
                plt.colorbar(
                    ax,
                    ax=[self.axes[subplot_index]],
                    label='Wind Speed (' + self._ds[dsname][u_field].attrs['units'] + ')',
                )

            else:
                self.axes[subplot_index].barbs(
                    xdata[::barb_step_x],
                    ydata[::barb_step_x],
                    u[::barb_step_x],
                    v[::barb_step_x],
                    **kwargs,
                )
            self.axes[subplot_index].set_yticks([])

        else:
            if 'cmap' in kwargs.keys():
                map_color = np.sqrt(
                    np.power(u[::barb_step_x, ::barb_step_y], 2)
                    + np.power(v[::barb_step_x, ::barb_step_y], 2)
                )
                map_color[np.isnan(map_color)] = 0
                ax = self.axes[subplot_index].barbs(
                    xdata[::barb_step_x, ::barb_step_y],
                    ydata[::barb_step_x, ::barb_step_y],
                    u[::barb_step_x, ::barb_step_y],
                    v[::barb_step_x, ::barb_step_y],
                    map_color,
                    **kwargs,
                )
                plt.colorbar(
                    ax,
                    ax=[self.axes[subplot_index]],
                    label='Wind Speed (' + self._ds[dsname][u_field].attrs['units'] + ')',
                )
            else:
                ax = self.axes[subplot_index].barbs(
                    xdata[::barb_step_x, ::barb_step_y],
                    ydata[::barb_step_x, ::barb_step_y],
                    u[::barb_step_x, ::barb_step_y],
                    v[::barb_step_x, ::barb_step_y],
                    **kwargs,
                )

        if day_night_background is True:
            self.day_night_background(subplot_index=subplot_index, dsname=dsname)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )

        self.axes[subplot_index].set_title(set_title)

        # Set YTitle
        if 'ytitle' in locals():
            self.axes[subplot_index].set_ylabel(ytitle)

        # Set X Limit - We want the same time axes for all subplots
        time_rng = [xdata.min(), xdata.max()]
        self.set_xrng(time_rng, subplot_index)

        # Set Y Limit
        if hasattr(self, 'yrng'):
            # Make sure that the yrng is not just the default
            if not np.all(self.yrng[subplot_index] == 0):
                self.set_yrng(self.yrng[subplot_index], subplot_index)
            else:
                if ydata is None:
                    our_data = xdata
                else:
                    our_data = ydata

                if np.isfinite(our_data).any():
                    if invert_y_axis is False:
                        yrng = [np.nanmin(our_data), np.nanmax(our_data)]
                    else:
                        yrng = [np.nanmax(our_data), np.nanmin(our_data)]
                else:
                    yrng = [0, 1]
                self.set_yrng(yrng, subplot_index)

        # Set X Format
        if len(subplot_index) == 1:
            days = self.xrng[subplot_index, 1] - self.xrng[subplot_index, 0]
        else:
            days = (
                self.xrng[subplot_index[0], subplot_index[1], 1]
                - self.xrng[subplot_index[0], subplot_index[1], 0]
            )

        # Put on an xlabel, but only if we are making the bottom-most plot
        if subplot_index[0] == self.axes.shape[0] - 1:
            self.axes[subplot_index].set_xlabel('Time [UTC]')

        myFmt = common.get_date_format(days)
        self.axes[subplot_index].xaxis.set_major_formatter(myFmt)
        return self.axes[subplot_index]

    def plot_time_height_xsection_from_1d_data(
        self,
        data_field,
        pres_field,
        dsname=None,
        subplot_index=(0,),
        set_title=None,
        day_night_background=False,
        num_time_periods=20,
        num_y_levels=20,
        invert_y_axis=True,
        cbar_label=None,
        set_shading='auto',
        **kwargs,
    ):
        """
        This will plot a time-height cross section from 1D datasets using
        nearest neighbor interpolation on a regular time by height grid.
        All that is needed are a data variable and a height variable.

        Parameters
        ----------
        data_field : str
            The name of the field to plot.
        pres_field : str
            The name of the height or pressure field to plot.
        dsname : str or None
            The name of the datastream to plot
        subplot_index : 2-tuple
            The index of the subplot to create the plot on.
        set_title : str or None
            The title of the plot.
        day_night_background : bool
            Set to true to plot the day/night background.
        num_time_periods : int
            Set to determine how many time periods. Setting to None
            will do one time period per day.
        num_y_levels : int
            The number of levels in the y axis to use.
        invert_y_axis : bool
            Set to true to invert the y-axis (recommended for
            pressure coordinates).
        cbar_label : str
            Option to overwrite default colorbar label.
        set_shading : string
            Option to to set the matplotlib.pcolormesh shading parameter.
            Default to 'auto'
        **kwargs : keyword arguments
            Additional keyword arguments will be passed
            into :func:`plt.pcolormesh`

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle pointing to the plot.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2'
                'or more datasets in the TimeSeriesDisplay'
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        dim = list(self._ds[dsname][data_field].dims)
        if len(dim) > 1:
            raise ValueError(
                'plot_time_height_xsection_from_1d_data only '
                'supports 1-D datasets. For datasets with 2 or '
                'more dimensions use plot().'
            )

        # Get data and dimensions
        data = self._ds[dsname][data_field].values
        xdata = self._ds[dsname][dim[0]].values

        # What we will do here is do a nearest-neighbor interpolation for each
        # member of the series. Coordinates are time, pressure
        pres = self._ds[dsname][pres_field]
        u_interp = NearestNDInterpolator((xdata, pres.values), data, rescale=True)
        # Mask points where we have no data
        # Count number of unique days
        x_times = pd.date_range(xdata.min(), xdata.max(), periods=num_time_periods)
        y_levels = np.linspace(np.nanmin(pres), np.nanmax(pres), num_y_levels)
        tdata, ydata = np.meshgrid(x_times, y_levels, indexing='ij')
        data = u_interp(tdata, ydata)
        ytitle = ''.join(['(', pres.attrs['units'], ')'])
        units = data_field + ' (' + self._ds[dsname][data_field].attrs['units'] + ')'

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        mesh = self.axes[subplot_index].pcolormesh(
            x_times, y_levels, np.transpose(data), shading=set_shading, **kwargs
        )

        if day_night_background is True:
            self.day_night_background(subplot_index=subplot_index, dsname=dsname)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )

        self.axes[subplot_index].set_title(set_title)

        # Set YTitle
        if 'ytitle' in locals():
            self.axes[subplot_index].set_ylabel(ytitle)

        # Set X Limit - We want the same time axes for all subplots
        time_rng = [x_times[0], x_times[-1]]

        self.set_xrng(time_rng, subplot_index)

        # Set Y Limit
        if hasattr(self, 'yrng'):
            # Make sure that the yrng is not just the default
            if not np.all(self.yrng[subplot_index] == 0):
                self.set_yrng(self.yrng[subplot_index], subplot_index)
            else:
                if ydata is None:
                    our_data = data.values
                else:
                    our_data = ydata
                if np.isfinite(our_data).any():
                    if invert_y_axis is False:
                        yrng = [np.nanmin(our_data), np.nanmax(our_data)]
                    else:
                        yrng = [np.nanmax(our_data), np.nanmin(our_data)]
                else:
                    yrng = [0, 1]
                self.set_yrng(yrng, subplot_index)

        # Set X Format
        if len(subplot_index) == 1:
            days = self.xrng[subplot_index, 1] - self.xrng[subplot_index, 0]
        else:
            days = (
                self.xrng[subplot_index[0], subplot_index[1], 1]
                - self.xrng[subplot_index[0], subplot_index[1], 0]
            )

        # Put on an xlabel, but only if we are making the bottom-most plot
        if subplot_index[0] == self.axes.shape[0] - 1:
            self.axes[subplot_index].set_xlabel('Time [UTC]')

        if ydata is not None:
            if cbar_label is None:
                self.add_colorbar(mesh, title=units, subplot_index=subplot_index)
            else:
                self.add_colorbar(mesh, title=cbar_label, subplot_index=subplot_index)
        myFmt = common.get_date_format(days)
        self.axes[subplot_index].xaxis.set_major_formatter(myFmt)

        return self.axes[subplot_index]

    def time_height_scatter(
        self,
        data_field=None,
        dsname=None,
        cmap='rainbow',
        alt_label=None,
        alt_field='alt',
        cb_label=None,
        **kwargs,
    ):
        """
        Create a time series plot of altitude and data variable with
        color also indicating value with a color bar. The Color bar is
        positioned to serve both as the indicator of the color intensity
        and the second y-axis.

        Parameters
        ----------
        data_field : str
            Name of data field in the dataset to plot on second y-axis.
        height_field : str
            Name of height field in the dataset to plot on first y-axis.
        dsname : str or None
            The name of the datastream to plot.
        cmap : str
            Colorbar color map to use.
        alt_label : str
            Altitude first y-axis label to use. If None, will try to use
            long_name and units.
        alt_field : str
            Label for field in the dataset to plot on first y-axis.
        cb_label : str
            Colorbar label to use. If not set will try to use
            long_name and units.
        **kwargs : keyword arguments
            Any other keyword arguments that will be passed
            into TimeSeriesDisplay.plot module when the figure
            is made.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Get data and dimensions
        data = self._ds[dsname][data_field]
        altitude = self._ds[dsname][alt_field]
        dim = list(self._ds[dsname][data_field].dims)
        xdata = self._ds[dsname][dim[0]]

        if alt_label is None:
            try:
                alt_label = altitude.attrs['long_name'] + ''.join(
                    [' (', altitude.attrs['units'], ')']
                )
            except KeyError:
                alt_label = alt_field

        if cb_label is None:
            try:
                cb_label = data.attrs['long_name'] + ''.join([' (', data.attrs['units'], ')'])
            except KeyError:
                cb_label = data_field

        colorbar_map = plt.cm.get_cmap(cmap)
        self.fig.subplots_adjust(left=0.1, right=0.86, bottom=0.16, top=0.91)
        ax1 = self.plot(alt_field, color='black', **kwargs)
        ax1.set_ylabel(alt_label)
        ax2 = ax1.twinx()
        sc = ax2.scatter(xdata.values, data.values, c=data.values, marker='.', cmap=colorbar_map)
        cbaxes = self.fig.add_axes(
            [
                self.fig.subplotpars.right + 0.02,
                self.fig.subplotpars.bottom,
                0.02,
                self.fig.subplotpars.top - self.fig.subplotpars.bottom,
            ]
        )
        cbar = plt.colorbar(sc, cax=cbaxes)
        ax2.set_ylim(cbar.mappable.get_clim())
        cbar.ax.set_ylabel(cb_label)
        ax2.set_yticklabels([])

        return self.axes[0]

    def qc_flag_block_plot(
        self,
        data_field=None,
        dsname=None,
        subplot_index=(0,),
        time_rng=None,
        assessment_color=None,
        edgecolor='face',
        set_shading='auto',
        cb_friendly=False,
        **kwargs,
    ):
        """
        Create a time series plot of embedded quality control values
        using broken barh plotting.

        Parameters
        ----------
        data_field : str
            Name of data field in the dataset to plot corresponding quality
            control.
        dsname : None or str
            If there is more than one datastream in the display object the
            name of the datastream needs to be specified. If set to None and
            there is only one datastream ACT will use the sole datastream
            in the object.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.
        time_rng : tuple or list
            List or tuple with (min, max) values to set the x-axis range limits.
        assessment_color : dict
            Dictionary lookup to override default assessment to color. Make sure
            assessment work is correctly set with case syntax.
        edgecolor : str or list
            Color name, list of color names or 'face' as defined in matplotlib.axes.Axes.broken_barh
        set_shading : string
            Option to to set the matplotlib.pcolormesh shading parameter.
            Default to 'auto'
        cb_friendly : boolean
            Set to true if you want to use the integrated colorblind friendly
            colors for green/red based on the Homeyer colormap
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.broken_barh`.

        """
        # Color to plot associated with assessment.
        color_lookup = {
            'Bad': 'red',
            'Incorrect': 'red',
            'Indeterminate': 'orange',
            'Suspect': 'orange',
            'Missing': 'darkgray',
            'Not Failing': 'green',
            'Acceptable': 'green',
        }
        if cb_friendly:
            color_lookup['Bad'] = (0.9285714285714286, 0.7130901016453677, 0.7130901016453677)
            color_lookup['Incorrect'] = (0.9285714285714286, 0.7130901016453677, 0.7130901016453677)
            color_lookup['Not Failing'] = (0.0, 0.4240129715562796, 0.4240129715562796),
            color_lookup['Acceptable'] = (0.0, 0.4240129715562796, 0.4240129715562796),

        if assessment_color is not None:
            for asses, color in assessment_color.items():
                color_lookup[asses] = color
                if asses == 'Incorrect':
                    color_lookup['Bad'] = color
                if asses == 'Suspect':
                    color_lookup['Indeterminate'] = color

        # Set up list of test names to use for missing values
        missing_val_long_names = [
            'Value equal to missing_value*',
            'Value set to missing_value*',
            'Value is equal to missing_value*',
            'Value is set to missing_value*',
        ]

        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Set up or get current plot figure
        if self.fig is None:
            self.fig = plt.figure()

        # Set up or get current axes
        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        ax = self.axes[subplot_index]

        # Set X Limit - We want the same time axes for all subplots
        data = self._ds[dsname][data_field]
        dim = list(self._ds[dsname][data_field].dims)
        xdata = self._ds[dsname][dim[0]]

        # Get data and attributes
        qc_data_field = self._ds[dsname].qcfilter.check_for_ancillary_qc(
            data_field, add_if_missing=False, cleanup=False
        )
        if qc_data_field is None:
            raise ValueError(f'No quality control ancillary variable in Dataset for {data_field}')

        flag_masks = self._ds[dsname][qc_data_field].attrs['flag_masks']
        flag_meanings = self._ds[dsname][qc_data_field].attrs['flag_meanings']
        flag_assessments = self._ds[dsname][qc_data_field].attrs['flag_assessments']

        # Get time ranges for green blocks
        time_delta = determine_time_delta(xdata.values)
        barh_list_green = reduce_time_ranges(xdata.values, time_delta=time_delta, broken_barh=True)

        # Set background to gray indicating not available data
        ax.set_facecolor('dimgray')

        # Check if plotting 2D data vs 1D data. 2D data will be summarized by
        # assessment category instead of showing each test.
        data_shape = self._ds[dsname][qc_data_field].shape
        if len(data_shape) > 1:
            cur_assessments = list(set(flag_assessments))
            cur_assessments.sort()
            cur_assessments.reverse()
            qc_data = np.full(data_shape, -1, dtype=np.int16)
            plot_colors = []
            tick_names = []

            index = self._ds[dsname][qc_data_field].values == 0
            if index.any():
                qc_data[index] = 0
                plot_colors.append(color_lookup['Not Failing'])
                tick_names.append('Not Failing')

            for ii, assess in enumerate(cur_assessments):
                if assess not in color_lookup:
                    color_lookup[assess] = list(mplcolors.CSS4_COLORS.keys())[ii]
                ii += 1
                assess_data = self._ds[dsname].qcfilter.get_masked_data(
                    data_field, rm_assessments=assess
                )

                if assess_data.mask.any():
                    qc_data[assess_data.mask] = ii
                    plot_colors.append(color_lookup[assess])
                    tick_names.append(assess)

            # Overwrite missing data. Not sure if we want to do this because VAPs set
            # the value to missing but the test is set to Bad. This tries to overcome that
            # by looking for correct test description that would only indicate the values
            # are missing not that they are set to missing by a test... most likely.
            missing_test_nums = []
            for ii, flag_meaning in enumerate(flag_meanings):
                # Check if the bit set is indicating missing data.
                for val in missing_val_long_names:
                    if re_search(val, flag_meaning):
                        test_num = parse_bit(flag_masks[ii])[0]
                        missing_test_nums.append(test_num)
            assess_data = self._ds[dsname].qcfilter.get_masked_data(
                data_field, rm_tests=missing_test_nums
            )
            if assess_data.mask.any():
                qc_data[assess_data.mask] = -1
                plot_colors.append(color_lookup['Missing'])
                tick_names.append('Missing')

            # Create a masked array to allow not plotting where values are missing
            qc_data = np.ma.masked_equal(qc_data, -1)

            dims = self._ds[dsname][qc_data_field].dims
            xvalues = self._ds[dsname][dims[0]].values
            yvalues = self._ds[dsname][dims[1]].values

            cMap = mplcolors.ListedColormap(plot_colors)
            mesh = ax.pcolormesh(
                xvalues,
                yvalues,
                np.transpose(qc_data),
                cmap=cMap,
                vmin=0,
                shading=set_shading,
            )
            divider = make_axes_locatable(ax)
            # Determine correct placement of words on colorbar
            tick_nums = (
                np.arange(0, len(tick_names) * 2 + 1) / (len(tick_names) * 2) * np.nanmax(qc_data)
            )[1::2]
            cax = divider.append_axes('bottom', size='5%', pad=0.3)
            cbar = self.fig.colorbar(
                mesh,
                cax=cax,
                orientation='horizontal',
                spacing='uniform',
                ticks=tick_nums,
                shrink=0.5,
            )
            cbar.ax.set_xticklabels(tick_names)

            # Set YTitle
            dim_name = list(set(self._ds[dsname][qc_data_field].dims) - {'time'})
            try:
                ytitle = f"{dim_name[0]} ({self._ds[dsname][dim_name[0]].attrs['units']})"
                ax.set_ylabel(ytitle)
            except KeyError:
                pass

            # Add which tests were set as text to the plot
            unique_values = []
            for ii in np.unique(self._ds[dsname][qc_data_field].values):
                unique_values.extend(parse_bit(ii))
            if len(unique_values) > 0:
                unique_values = list(set(unique_values))
                unique_values.sort()
                unique_values = [str(ii) for ii in unique_values]
                self.fig.text(
                    0.5,
                    -0.35,
                    f"QC Tests Tripped: {', '.join(unique_values)}",
                    transform=ax.transAxes,
                    horizontalalignment='center',
                    verticalalignment='center',
                    fontweight='bold',
                )

        else:

            test_nums = []
            for ii, assess in enumerate(flag_assessments):
                if assess not in color_lookup:
                    color_lookup[assess] = list(mplcolors.CSS4_COLORS.keys())[ii]

                # Plot green data first.
                ax.broken_barh(
                    barh_list_green,
                    (ii, ii + 1),
                    facecolors=color_lookup['Not Failing'],
                    edgecolor=edgecolor,
                    **kwargs,
                )

                # Get test number from flag_mask bitpacked number
                test_nums.append(parse_bit(flag_masks[ii]))
                # Get masked array data to use mask for finding if/where test is set
                data = self._ds[dsname].qcfilter.get_masked_data(
                    data_field, rm_tests=test_nums[-1]
                )
                if np.any(data.mask):
                    # Get time ranges from time and masked data
                    barh_list = reduce_time_ranges(
                        xdata.values[data.mask], time_delta=time_delta, broken_barh=True
                    )
                    # Check if the bit set is indicating missing data. If so change
                    # to different plotting color than what is in flag_assessments.
                    for val in missing_val_long_names:
                        if re_search(val, flag_meanings[ii]):
                            assess = 'Missing'
                            break
                    # Lay down blocks of tripped tests using correct color
                    ax.broken_barh(
                        barh_list,
                        (ii, ii + 1),
                        facecolors=color_lookup[assess],
                        edgecolor=edgecolor,
                        **kwargs,
                    )

                # Add test description to plot.
                ax.text(xdata.values[0], ii + 0.5, ' ' + flag_meanings[ii], va='center')

            # Change y ticks to test number
            plt.yticks(
                [ii + 0.5 for ii in range(0, len(test_nums))],
                labels=['Test ' + str(ii[0]) for ii in test_nums],
            )
            # Set ylimit to number of tests plotted
            ax.set_ylim(0, len(flag_assessments))

        # Set X Limit - We want the same time axes for all subplots
        if not hasattr(self, 'time_rng'):
            if time_rng is not None:
                self.time_rng = list(time_rng)
            else:
                self.time_rng = [xdata.min().values, xdata.max().values]

        self.set_xrng(self.time_rng, subplot_index)

        # Get X format - We want the same time axes for all subplots
        if hasattr(self, 'time_fmt'):
            ax.xaxis.set_major_formatter(self.time_fmt)
        else:
            # Set X Format
            if len(subplot_index) == 1:
                days = self.xrng[subplot_index, 1] - self.xrng[subplot_index, 0]
            else:
                days = (
                    self.xrng[subplot_index[0], subplot_index[1], 1]
                    - self.xrng[subplot_index[0], subplot_index[1], 0]
                )

            myFmt = common.get_date_format(days)
            ax.xaxis.set_major_formatter(myFmt)
            self.time_fmt = myFmt

        return self.axes[subplot_index]

    def fill_between(
        self,
        field,
        dsname=None,
        subplot_index=(0,),
        set_title=None,
        secondary_y=False,
        **kwargs,
    ):
        """
        Makes a fill_between plot, based on matplotlib

        Parameters
        ----------
        field : str
            The name of the field to plot.
        dsname : None or str
            If there is more than one datastream in the display object the
            name of the datastream needs to be specified. If set to None and
            there is only one datastream ACT will use the sole datastream
            in the object.
        subplot_index : 1 or 2D tuple, list, or array
            The index of the subplot to set the x range of.
        set_title : str
            The title for the plot.
        secondary_y : boolean
            Option to indicate if the data should be plotted on second y-axis.
        **kwargs : keyword arguments
            The keyword arguments for :func:`plt.plot` (1D timeseries) or
            :func:`plt.pcolormesh` (2D timeseries).

        Returns
        -------
        ax : matplotlib axis handle
            The matplotlib axis handle of the plot.

        """
        if dsname is None and len(self._ds.keys()) > 1:
            raise ValueError(
                'You must choose a datastream when there are 2 '
                'or more datasets in the TimeSeriesDisplay '
                'object.'
            )
        elif dsname is None:
            dsname = list(self._ds.keys())[0]

        # Get data and dimensions
        data = self._ds[dsname][field]
        dim = list(self._ds[dsname][field].dims)
        xdata = self._ds[dsname][dim[0]]

        if 'units' in data.attrs:
            ytitle = ''.join(['(', data.attrs['units'], ')'])
        else:
            ytitle = field

        # Get the current plotting axis, add day/night background and plot data
        if self.fig is None:
            self.fig = plt.figure()

        if self.axes is None:
            self.axes = np.array([plt.axes()])
            self.fig.add_axes(self.axes[0])

        # Set ax to appropriate axis
        if secondary_y is False:
            ax = self.axes[subplot_index]
        else:
            ax = self.axes[subplot_index].twinx()

        ax.fill_between(xdata.values, data, **kwargs)

        # Set X Format
        if len(subplot_index) == 1:
            days = self.xrng[subplot_index, 1] - self.xrng[subplot_index, 0]
        else:
            days = (
                self.xrng[subplot_index[0], subplot_index[1], 1]
                - self.xrng[subplot_index[0], subplot_index[1], 0]
            )

        myFmt = common.get_date_format(days)
        ax.xaxis.set_major_formatter(myFmt)

        # Set X format - We want the same time axes for all subplots
        if not hasattr(self, 'time_fmt'):
            self.time_fmt = myFmt

        # Put on an xlabel, but only if we are making the bottom-most plot
        if subplot_index[0] == self.axes.shape[0] - 1:
            self.axes[subplot_index].set_xlabel('Time [UTC]')

        # Set YTitle
        ax.set_ylabel(ytitle)

        # Set Title
        if set_title is None:
            set_title = ' '.join(
                [
                    dsname,
                    field,
                    'on',
                    dt_utils.numpy_to_arm_date(self._ds[dsname].time.values[0]),
                ]
            )
        if secondary_y is False:
            ax.set_title(set_title)

        return self.axes[subplot_index]



"""
ARM Data Surveyor (ADS)
Command line wrapper around ACT.  Not all
features of ACT are included as options in ADS.
Please see the examples.txt for examples on how
to use ADS.

Author: Jason Hemedinger

"""

import argparse
import re
import json
import glob
import ast
import pathlib
import matplotlib.pyplot as plt
import numpy as np
import act

try:
    import cartopy.crs as ccrs
    CARTOPY_AVAILABLE = True
except ImportError:
    CARTOPY_AVAILABLE = False


def option_error_check(args, error_fields, check_all=False):
    '''
    This will check the args object for keys to see if they are set. If
    at least one key is not set will or all keys wiht check_all set
    will print error message and exit.
    '''

    if not isinstance(error_fields, (list, tuple)):
        error_fields = [error_fields]

    print_error = False
    how_many = 'one'
    if check_all is False and not any([vars(args)[ii] for ii in error_fields]):
        print_error = True

    if check_all is True and not all([vars(args)[ii] for ii in error_fields]):
        print_error = True
        how_many = 'all'

    if print_error:
        prepend = '--'
        for ii, value in enumerate(error_fields):
            if not value.startswith(prepend):
                error_fields[ii] = prepend + value

        print(f"\n{pathlib.Path(__file__).name}: error: {how_many} of the arguments "
              f"{' '.join(error_fields)} is requried\n")
        exit()


def find_drop_vars(args):
    '''
    This will check if more than one file is to be read. If so read one file
    and get list of variables to not read based on the fields arguments and
    corresponding QC or dimention variables. This will significantly speed up
    the reading time for reading many files.
    '''
    files = glob.glob(args.file_path)
    drop_vars = []
    if len(files) > 1:
        ds = act.io.armfiles.read_netcdf(files[0])
        ds.clean.cleanup()
        drop_vars = set(ds.data_vars)
        keep_vars = ['latitude', 'longitude']
        if args.field is not None:
            keep_vars.append(args.field)

        if args.fields is not None:
            keep_vars.extend(set(args.fields))

        if args.wind_fields is not None:
            keep_vars.extend(set(args.wind_fields))

        if args.station_fields is not None:
            keep_vars.extend(set(args.station_fields))

        if args.latitude is not None:
            keep_vars.append(args.latitude)

        if args.longitude is not None:
            keep_vars.append(args.longitude)

        if args.x_field is not None:
            keep_vars.append(args.x_field)

        if args.y_field is not None:
            keep_vars.append(args.y_field)

        if args.u_wind is not None:
            keep_vars.append(args.u_wind)

        if args.v_wind is not None:
            keep_vars.append(args.v_wind)

        if args.p_field is not None:
            keep_vars.append(args.p_field)

        if args.t_field is not None:
            keep_vars.append(args.t_field)

        if args.td_field is not None:
            keep_vars.append(args.td_field)

        if args.spd_field is not None:
            keep_vars.append(args.spd_field)

        if args.dir_field is not None:
            keep_vars.append(args.dir_field)

        keep_vars_additional = []
        for var_name in keep_vars:
            qc_var_name = ds.qcfilter.check_for_ancillary_qc(
                var_name, add_if_missing=False, cleanup=False)
            if qc_var_name is not None:
                keep_vars_additional.append(qc_var_name)

            try:
                keep_vars_additional.extend(ds[var_name].dims)
            except KeyError:
                pass

        drop_vars = drop_vars - set(keep_vars) - set(keep_vars_additional)

    return drop_vars


def geodisplay(args):
    ds = act.io.armfiles.read_netcdf(args.file_path)

    dsname = args.dsname
    if dsname == _default_dsname:
        try:
            dsname = ds.attrs['datastream']
        except KeyError:
            pass

    display = act.plotting.GeographicPlotDisplay({dsname: ds},
                                                 figsize=args.figsize)

    display.geoplot(data_field=args.field, lat_field=args.latitude,
                    lon_field=args.longitude, dsname=dsname,
                    cbar_label=args.cb_label, title=args.set_title,
                    plot_buffer=args.plot_buffer, stamen=args.stamen,
                    tile=args.tile, cartopy_feature=args.cfeatures,
                    cmap=args.cmap, text=args.text, gridlines=args.gridlines,
                    projection=args.projection, **args.kwargs)

    plt.savefig(args.out_path)
    plt.show()
    plt.close(display.fig)

    ds.close()


def skewt(args):
    ds = act.io.armfiles.read_netcdf(args.file_path)

    subplot_index = args.subplot_index

    dsname = args.dsname
    if dsname == _default_dsname:
        try:
            dsname = ds.attrs['datastream']
        except KeyError:
            pass

    display = act.plotting.SkewTDisplay({dsname: ds}, figsize=args.figsize)

    if args.from_u_and_v:
        display.plot_from_u_and_v(u_field=args.u_wind, v_field=args.v_wind,
                                  p_field=args.p_field, t_field=args.t_field,
                                  td_field=args.td_field,
                                  subplot_index=subplot_index,
                                  dsname=dsname, show_parcel=args.show_parcel,
                                  p_levels_to_plot=args.plevels_plot,
                                  shade_cape=args.shade_cape,
                                  shade_cin=args.shade_cin,
                                  set_title=args.set_title,
                                  plot_barbs_kwargs=args.plot_barbs_kwargs,
                                  plot_kwargs=args.plot_kwargs)

    if args.from_spd_and_dir:
        display.plot_from_spd_and_dir(spd_field=args.spd_field,
                                      dir_field=args.dir_field,
                                      p_field=args.p_field,
                                      t_field=args.t_field,
                                      td_field=args.td_field,
                                      dsname=dsname,
                                      **args.kwargs)

    plt.savefig(args.out_path)
    plt.show()
    plt.close(display.fig)

    ds.close()


def xsection(args):
    ds = act.io.armfiles.read_netcdf(args.file_path)

    subplot_index = args.subplot_index

    dsname = args.dsname
    if dsname == _default_dsname:
        try:
            dsname = ds.attrs['datastream']
        except KeyError:
            pass

    display = act.plotting.XSectionDisplay({dsname: ds}, figsize=args.figsize)

    if args.plot_xsection:
        display.plot_xsection(dsname=dsname, varname=args.field,
                              x=args.x_field, y=args.y_field,
                              subplot_index=subplot_index,
                              sel_kwargs=args.sel_kwargs,
                              isel_kwargs=args.isel_kwargs, **args.kwargs)

    if args.xsection_map:
        display.plot_xsection_map(dsname=dsname, varname=args.field,
                                  subplot_index=subplot_index,
                                  coastlines=args.coastlines,
                                  background=args.background,
                                  **args.kwargs)

    plt.savefig(args.out_path)
    plt.show()
    plt.close(display.fig)

    ds.close()


def wind_rose(args):

    drop_vars = find_drop_vars(args)

    ds = act.io.armfiles.read_netcdf(args.file_path, drop_variables=drop_vars)

    subplot_index = args.subplot_index

    dsname = args.dsname
    if dsname == _default_dsname:
        try:
            dsname = ds.attrs['datastream']
        except KeyError:
            pass

    display = act.plotting.WindRoseDisplay({dsname: ds},
                                           figsize=args.figsize)

    display.plot(dir_field=args.dir_field, spd_field=args.spd_field,
                 subplot_index=subplot_index,
                 dsname=dsname, cmap=args.cmap,
                 set_title=args.set_title,
                 num_dirs=args.num_dir, spd_bins=args.spd_bins,
                 tick_interval=args.tick_interval, **args.kwargs)
    plt.savefig(args.out_path)
    plt.show()
    plt.close(display.fig)

    ds.close()


def timeseries(args):

    drop_vars = find_drop_vars(args)

    ds = act.io.armfiles.read_netcdf(args.file_path, drop_variables=drop_vars)

    if args.cleanup:
        ds.clean.cleanup()

    subplot_shape = args.subplot_shape
    subplot_index = args.subplot_index

    dsname = args.dsname
    if dsname == _default_dsname:
        try:
            dsname = ds.attrs['datastream']
        except KeyError:
            pass

    display = act.plotting.TimeSeriesDisplay(
        {dsname: ds}, figsize=args.figsize,
        subplot_shape=subplot_shape)

    options = ['plot', 'barbs_spd_dir', 'barbs_u_v', 'xsection_from_1d',
               'time_height_scatter', 'qc', 'fill_between', 'multi_panel']
    option_error_check(args, options)

    if args.plot:
        option_error_check(args, 'field')
        if args.set_yrange is not None:
            yrange = list(map(float, args.set_yrange))
        else:
            yrange = args.set_yrange
        display.plot(
            field=args.field, dsname=dsname, cmap=args.cmap,
            set_title=args.set_title, add_nan=args.add_nan,
            subplot_index=subplot_index,
            use_var_for_y=args.var_y,
            day_night_background=args.day_night,
            invert_y_axis=args.invert_y_axis,
            abs_limits=args.abs_limits, time_rng=args.time_rng,
            assessment_overplot=args.assessment_overplot,
            assessment_overplot_category=args.overplot_category,
            assessment_overplot_category_color=args.category_color,
            force_line_plot=args.force_line_plot, labels=args.labels,
            cbar_label=args.cb_label, secondary_y=args.secondary_y,
            y_rng=yrange,
            **args.kwargs)

    if args.barbs_spd_dir:
        display.plot_barbs_from_spd_dir(
            dir_field=args.dir_field,
            spd_field=args.spd_field,
            pres_field=args.p_field,
            dsname=dsname,
            **args.kwargs)

    if args.barbs_u_v:
        display.plot_barbs_from_u_v(
            u_field=args.u_wind, v_field=args.v_wind,
            pres_field=args.p_field, dsname=dsname,
            set_title=args.set_title,
            invert_y_axis=args.invert_y_axis,
            day_night_background=args.day_night,
            num_barbs_x=args.num_barb_x,
            num_barbs_y=args.num_barb_y,
            use_var_for_y=args.var_y,
            subplot_index=subplot_index,
            **args.kwargs)

    if args.xsection_from_1d:
        option_error_check(args, 'field')

        display.plot_time_height_xsection_from_1d_data(
            data_field=args.field, pres_field=args.p_field,
            dsname=dsname, set_title=args.set_title,
            day_night_background=args.day_night,
            num_time_periods=args.num_time_periods,
            num_y_levels=args.num_y_levels,
            invert_y_axis=args.invert_y_axis,
            subplot_index=subplot_index,
            cbar_label=args.cb_label,
            **args.kwargs)

    if args.time_height_scatter:
        option_error_check(args, 'field')

        display.time_height_scatter(
            data_field=args.field, dsname=dsname,
            cmap=args.cmap, alt_label=args.alt_label,
            alt_field=args.alt_field, cb_label=args.cb_label,
            **args.kwargs)

    if args.qc:
        option_error_check(args, 'field')
        display.qc_flag_block_plot(
            data_field=args.field, dsname=dsname,
            subplot_index=subplot_index,
            time_rng=args.time_rng,
            assessment_color=args.assessment_color,
            **args.kwargs)

    if args.fill_between:
        option_error_check(args, 'field')

        display.fill_between(
            field=args.field, dsname=dsname,
            subplot_index=subplot_index,
            set_title=args.set_title,
            secondary_y=args.secondary_y,
            **args.kwargs)

    if args.multi_panel:
        option_error_check(args, ['fields', 'plot_type'], check_all=True)

        for i, j, k in zip(args.fields, subplot_index, args.plot_type):
            if k == 'plot':
                display.plot(
                    field=i, dsname=dsname, cmap=args.cmap,
                    set_title=args.set_title, add_nan=args.add_nan,
                    subplot_index=j,
                    use_var_for_y=args.var_y,
                    day_night_background=args.day_night,
                    invert_y_axis=args.invert_y_axis,
                    abs_limits=args.abs_limits, time_rng=args.time_rng,
                    assessment_overplot=args.assessment_overplot,
                    assessment_overplot_category=args.overplot_category,
                    assessment_overplot_category_color=args.category_color,
                    force_line_plot=args.force_line_plot, labels=args.labels,
                    cbar_label=args.cb_label, secondary_y=args.secondary_y,
                    **args.kwargs)

            if k == 'qc':
                display.qc_flag_block_plot(
                    data_field=i, dsname=dsname,
                    subplot_index=j,
                    time_rng=args.time_rng,
                    assessment_color=args.assessment_color,
                    **args.kwargs)

    plt.savefig(args.out_path)
    plt.show()
    plt.close(display.fig)

    ds.close()


def histogram(args):

    drop_vars = find_drop_vars(args)

    ds = act.io.armfiles.read_netcdf(args.file_path, drop_variables=drop_vars)

    subplot_shape = args.subplot_shape
    subplot_index = args.subplot_index

    dsname = args.dsname
    if dsname == _default_dsname:
        try:
            dsname = ds.attrs['datastream']
        except KeyError:
            pass

    display = act.plotting.HistogramDisplay(
        {dsname: ds}, figsize=args.figsize,
        subplot_shape=subplot_shape)

    if args.stacked_bar_graph:
        display.plot_stacked_bar_graph(
            field=args.field, dsname=dsname,
            bins=args.bins, density=args.density,
            sortby_field=args.sortby_field,
            sortby_bins=args.sortby_bins,
            set_title=args.set_title,
            subplot_index=subplot_index,
            **args.kwargs)

    if args.size_dist:
        display.plot_size_distribution(
            field=args.field, bins=args.bin_field,
            time=args.time, dsname=dsname,
            set_title=args.set_title,
            subplot_index=subplot_index,
            **args.kwargs)

    if args.stairstep:
        display.plot_stairstep_graph(
            field=args.field, dsname=dsname,
            bins=args.bins, density=args.density,
            sortby_field=args.sortby_field,
            sortby_bins=args.sortby_bins,
            set_title=args.set_title,
            subplot_index=subplot_index,
            **args.kwargs)

    if args.heatmap:
        display.plot_heatmap(
            x_field=args.x_field, y_field=args.y_field,
            dsname=dsname, x_bins=args.x_bins,
            y_bins=args.y_bins, set_title=args.set_title,
            density=args.density,
            subplot_index=subplot_index, **args.kwargs)

    plt.savefig(args.out_path)
    plt.show()
    plt.close(display.fig)

    ds.close()


def contour(args):
    files = glob.glob(args.file_path)
    files.sort()

    time = args.time
    data = {}
    fields = {}
    wind_fields = {}
    station_fields = {}
    for f in files:
        ds = act.io.armfiles.read_netcdf(f)
        data.update({f: ds})
        fields.update({f: args.fields})
        wind_fields.update({f: args.wind_fields})
        station_fields.update({f: args.station_fields})

    display = act.plotting.ContourDisplay(data, figsize=args.figsize)

    if args.create_contour:
        display.create_contour(fields=fields, time=time, function=args.function,
                               grid_delta=args.grid_delta,
                               grid_buffer=args.grid_buffer,
                               subplot_index=args.subplot_index,
                               **args.kwargs)

    if args.contourf:
        display.contourf(x=args.x, y=args.y, z=args.z,
                         subplot_index=args.subplot_index,
                         **args.kwargs)

    if args.plot_contour:
        display.contour(x=args.x, y=args.y, z=args.z,
                        subplot_index=args.subplot_index,
                        **args.kwargs)

    if args.vectors_spd_dir:
        display.plot_vectors_from_spd_dir(fields=wind_fields, time=time,
                                          mesh=args.mesh, function=args.function,
                                          grid_delta=args.grid_delta,
                                          grid_buffer=args.grid_buffer,
                                          subplot_index=args.subplot_index,
                                          **args.kwargs)

    if args.barbs:
        display.barbs(x=args.x, y=args.y, u=args.u, v=args.v,
                      subplot_index=args.subplot_index,
                      **args.kwargs)

    if args.plot_station:
        display.plot_station(fields=station_fields, time=time,
                             text_color=args.text_color,
                             subplot_index=args.subplot_index,
                             **args.kwargs)

    plt.savefig(args.out_path)
    plt.show()
    plt.close(display.fig)

    ds.close()


# Define new funciton for argparse to allow specific rules for
# parsing files containing arguments. This works by this function being
# called for each line in the configuration file.
def convert_arg_line_to_args(line):
    for arg in line.split():
        if not arg.strip():  # If empty line or only white space skip
            continue
        if arg.startswith('#'):  # If line starts with comment skip
            break
        yield arg


def main():
    prefix_char = '@'
    parser = argparse.ArgumentParser(
        description=(f'Create plot from a data file. Can use command line opitons '
                     f'or point to a configuration file using {prefix_char} character.'))

    # Allow user to reference a file by using the @ symbol for a specific
    # argument value
    parser = argparse.ArgumentParser(fromfile_prefix_chars=prefix_char)

    # Update the file parsing logic to skip commented lines
    parser.convert_arg_line_to_args = convert_arg_line_to_args

    parser.add_argument('-f', '--file_path', type=str, required=True,
                        help=('Required: Full path to file for creating Plot. For multiple '
                              'files use terminal syntax for matching muliple files. '
                              'For example "sgpmetE13.b1.202007*.*.nc" will match all files '
                              'for the month of July in 2020. Need to use double quotes '
                              'to stop terminal from expanding the search, and let the '
                              'python program perform search.'))
    out_path_default = 'image.png'
    parser.add_argument('-o', '--out_path', type=str, default=out_path_default,
                        help=("Full path filename to use for saving image. "
                              "Default is '{out_path_default}'. If only a path is given "
                              "will use that path with image name '{out_path_default}', "
                              "else will use filename given."))
    parser.add_argument('-fd', '--field', type=str, default=None,
                        help='Name of the field to plot')
    parser.add_argument('-fds', '--fields', nargs='+',
                        type=str, default=None,
                        help='Name of the fields to use to plot')
    parser.add_argument('-wfs', '--wind_fields', nargs='+',
                        type=str, default=None,
                        help='Wind field names used to plot')
    parser.add_argument('-sfs', '--station_fields', nargs='+',
                        type=str, default=None,
                        help='Station field names to plot sites')
    default = 'lat'
    parser.add_argument('-lat', '--latitude', type=str, default=default,
                        help=f"Name of latitude variable in file. Default is '{default}'")
    default = 'lon'
    parser.add_argument('-lon', '--longitude', type=str, default=default,
                        help=f"Name of longitude variable in file. Default is '{default}'")
    parser.add_argument('-xf', '--x_field', type=str, default=None,
                        help='Name of variable to plot on x axis')
    parser.add_argument('-yf', '--y_field', type=str, default=None,
                        help='Name of variable to plot on y axis')
    parser.add_argument('-x', type=np.array,
                        help='x coordinates or grid for z')
    parser.add_argument('-y', type=np.array,
                        help='y coordinates or grid for z')
    parser.add_argument('-z', type=np.array,
                        help='Values over which to contour')
    default = 'u_wind'
    parser.add_argument('-u', '--u_wind', type=str, default=default,
                        help=f"File variable name for u_wind wind component. Default is '{default}'")
    default = 'v_wind'
    parser.add_argument('-v', '--v_wind', type=str, default=default,
                        help=f"File variable name for v_wind wind compenent. Default is '{default}'")
    default = 'pres'
    parser.add_argument('-pf', '--p_field', type=str, default=default,
                        help=f"File variable name for pressure. Default is '{default}'")
    default = 'tdry'
    parser.add_argument('-tf', '--t_field', type=str, default=default,
                        help=f"File variable name for temperature. Default is '{default}'")
    default = 'dp'
    parser.add_argument('-tdf', '--td_field', type=str, default=default,
                        help=f"File variable name for dewpoint temperature. Default is '{default}'")
    default = 'wspd'
    parser.add_argument('-sf', '--spd_field', type=str, default=default,
                        help=f"File variable name for wind speed. Default is '{default}'")
    default = 'deg'
    parser.add_argument('-df', '--dir_field', type=str, default=default,
                        help=f"File variable name for wind direction. Default is '{default}'")
    parser.add_argument('-al', '--alt_label', type=str, default=None,
                        help='Altitude axis label')
    default = 'alt'
    parser.add_argument('-af', '--alt_field', type=str, default=default,
                        help=f"File variable name for altitude. Default is '{default}'")
    global _default_dsname
    _default_dsname = 'act_datastream'
    parser.add_argument('-ds', '--dsname', type=str, default=_default_dsname,
                        help=f"Name of datastream to plot. Default is '{_default_dsname}'")
    default = '(0, )'
    parser.add_argument('-si', '--subplot_index', type=ast.literal_eval,
                        default=default,
                        help=f'Index of the subplot via tuple syntax. '
                        f'Example for two plots is "(0,), (1,)". '
                        f"Default is '{default}'")
    default = (1, )
    parser.add_argument('-ss', '--subplot_shape', nargs='+', type=int,
                        default=default,
                        help=(f'The number of (rows, columns) '
                              f'for the subplots in the display. '
                              f'Default is {default}'))
    plot_type_options = ['plot', 'qc']
    parser.add_argument('-pt', '--plot_type', nargs='+', type=str,
                        help=f'Type of plot to make. Current options include: '
                        f'{plot_type_options}')
    parser.add_argument('-vy', '--var_y', type=str, default=None,
                        help=('Set this to the name of a data variable in '
                              'the Dataset to use as the y-axis variable '
                              'instead of the default dimension.'))
    parser.add_argument('-plp', '--plevels_plot',
                        type=np.array, default=None,
                        help='Pressure levels to plot the wind barbs on.')
    parser.add_argument('-cbl', '--cb_label', type=str, default=None,
                        help='Colorbar label to use')
    parser.add_argument('-st', '--set_title', type=str, default=None,
                        help='Title for the plot')
    default = 0.08
    parser.add_argument('-pb', '--plot_buffer', type=float, default=default,
                        help=(f'Buffer to add around data on plot in lat '
                              f'and lon dimension. Default is {default}'))
    default = 'terrain-background'
    parser.add_argument('-sm', '--stamen', type=str, default=default,
                        help=f"Dataset to use for background image. Default is '{default}'")
    default = 8
    parser.add_argument('-tl', '--tile', type=int, default=default,
                        help=f'Tile zoom to use with background image. Default is {default}')
    parser.add_argument('-cfs', '--cfeatures', nargs='+', type=str, default=None,
                        help='Cartopy feature to add to plot')
    parser.add_argument('-txt', '--text', type=json.loads, default=None,
                        help=('Dictionary of {text:[lon,lat]} to add to plot. '
                              'Can have more than one set of text to add.'))
    default = 'rainbow'
    parser.add_argument('-cm', '--cmap', default=default,
                        help=f"colormap to use. Defaut is '{default}'")
    parser.add_argument('-abl', '--abs_limits', nargs='+', type=float,
                        default=(None, None),
                        help=('Sets the bounds on plot limits even if data '
                              'values exceed those limits. Y axis limits. Default is no limits.'))
    parser.add_argument('-tr', '--time_rng', nargs='+', type=float, default=None,
                        help=('List or tuple with (min,max) values to set the '
                              'x-axis range limits'))
    default = 20
    parser.add_argument('-nd', '--num_dir', type=int, default=default,
                        help=(f'Number of directions to splot the wind rose into. '
                              f'Default is {default}'))
    parser.add_argument('-sb', '--spd_bins', nargs='+', type=float, default=None,
                        help='Bin boundaries to sort the wind speeds into')
    default = 3
    parser.add_argument('-ti', '--tick_interval', type=int, default=default,
                        help=(f'Interval (in percentage) for the ticks '
                              f'on the radial axis. Default is {default}'))
    parser.add_argument('-ac', '--assessment_color', type=json.loads,
                        default=None,
                        help=('dictionary lookup to override default '
                              'assessment to color'))
    default = False
    parser.add_argument('-ao', '--assessment_overplot',
                        default=default, action='store_true',
                        help=(f'Option to overplot quality control colored '
                              f'symbols over plotted data using '
                              f'flag_assessment categories. Default is {default}'))
    default = {'Incorrect': ['Bad', 'Incorrect'],
               'Suspect': ['Indeterminate', 'Suspect']}
    parser.add_argument('-oc', '--overplot_category', type=json.loads, default=default,
                        help=(f'Look up to categorize assessments into groups. '
                              f'This allows using multiple terms for the same '
                              f'quality control level of failure. '
                              f'Also allows adding more to the defaults. Default is {default}'))
    default = {'Incorrect': 'red', 'Suspect': 'orange'}
    parser.add_argument('-co', '--category_color', type=json.loads,
                        default=default,
                        help=(f'Lookup to match overplot category color to '
                              f'assessment grouping. Default is {default}'))
    parser.add_argument('-flp', '--force_line_plot', default=False,
                        action='store_true',
                        help='Option to plot 2D data as 1D line plots')
    parser.add_argument('-l', '--labels', nargs='+', default=False,
                        type=str,
                        help=('Option to overwrite the legend labels. '
                              'Must have same dimensions as number of '
                              'lines plottes.'))
    parser.add_argument('-sy', '--secondary_y', default=False, action='store_true',
                        help='Option to plot on secondary y axis')
    if CARTOPY_AVAILABLE:
        default = ccrs.PlateCarree()
        parser.add_argument('-prj', '--projection', type=str,
                            default=default,
                            help=f"Projection to use on plot. Default is {default}")
    default = 20
    parser.add_argument('-bx', '--num_barb_x', type=int, default=default,
                        help=f'Number of wind barbs to plot in the x axis. Default is {default}')
    default = 20
    parser.add_argument('-by', '--num_barb_y', type=int, default=default,
                        help=f"Number of wind barbs to plot in the y axis. Default is {default}")
    default = 20
    parser.add_argument('-tp', '--num_time_periods', type=int, default=default,
                        help=f'Set how many time periods. Default is {default}')
    parser.add_argument('-bn', '--bins', nargs='+', type=int, default=None,
                        help='histogram bin boundaries to use')
    parser.add_argument('-bf', '--bin_field', type=str, default=None,
                        help=('name of the field that stores the '
                              'bins for the spectra'))
    parser.add_argument('-xb', '--x_bins', nargs='+', type=int, default=None,
                        help='Histogram bin boundaries to use for x axis variable')
    parser.add_argument('-yb', '--y_bins', nargs='+', type=int, default=None,
                        help='Histogram bin boundaries to use for y axis variable')
    parser.add_argument('-t', '--time', type=str, default=None,
                        help='Time period to be plotted')
    parser.add_argument('-sbf', '--sortby_field', type=str, default=None,
                        help='Sort histograms by a given field parameter')
    parser.add_argument('-sbb', '--sortby_bins', nargs='+', type=int,
                        default=None,
                        help='Bins to sort the histograms by')
    default = 20
    parser.add_argument('-nyl', '--num_y_levels', type=int, default=default,
                        help=f'Number of levels in the y axis to use. Default is {default}')
    parser.add_argument('-sk', '--sel_kwargs', type=json.loads, default=None,
                        help=('The keyword arguments to pass into '
                              ':py:func:`xarray.DataArray.sel`'))
    parser.add_argument('-ik', '--isel_kwargs', type=json.loads, default=None,
                        help=('The keyword arguments to pass into '
                              ':py:func:`xarray.DataArray.sel`'))
    default = 'cubic'
    parser.add_argument('-fn', '--function', type=str, default=default,
                        help=(f'Defaults to cubic function for interpolation. '
                              f'See scipy.interpolate.Rbf for additional options. '
                              f'Default is {default}'))
    default = 0.1
    parser.add_argument('-gb', '--grid_buffer', type=float, default=default,
                        help=f'Buffer to apply to grid. Default is {default}')
    default = (0.01, 0.01)
    parser.add_argument('-gd', '--grid_delta', nargs='+',
                        type=float, default=default,
                        help=f'X and Y deltas for creating grid. Default is {default}')
    parser.add_argument('-fg', '--figsize', nargs='+', type=float,
                        default=None,
                        help='Width and height in inches of figure')
    default = 'white'
    parser.add_argument('-tc', '--text_color', type=str, default=default,
                        help=f"Color of text. Default is '{default}'")
    parser.add_argument('-kwargs', type=json.loads, default=dict(),
                        help='keyword arguments to use in plotting function')
    parser.add_argument('-pk', '--plot_kwargs', type=json.loads, default=dict(),
                        help=("Additional keyword arguments to pass "
                              "into MetPy's SkewT.plot"))
    parser.add_argument('-pbk', '--plot_barbs_kwargs', type=json.loads,
                        default=dict(),
                        help=("Additional keyword arguments to pass "
                              "into MetPy's SkewT.plot_barbs"))
    default = True
    parser.add_argument('-cu', '--cleanup', default=default, action='store_false',
                        help=f'Turn off standard methods for obj cleanup. Default is {default}')
    parser.add_argument('-gl', '--gridlines', default=False, action='store_true',
                        help='Use latitude and lingitude gridlines.')
    parser.add_argument('-cl', '--coastlines', default=False, action='store_true',
                        help='Plot coastlines on geographical map')
    parser.add_argument('-bg', '--background', default=False, action='store_true',
                        help='Plot a stock image background')
    parser.add_argument('-nan', '--add_nan', default=False, action='store_true',
                        help='Fill in data gaps with NaNs')
    parser.add_argument('-dn', '--day_night', default=False, action='store_true',
                        help=("Fill in color coded background according "
                              "to time of day."))
    parser.add_argument('-yr', '--set_yrange', default=None, nargs=2,
                        help=("Set the yrange for the specific plot"))
    parser.add_argument('-iya', '--invert_y_axis', default=False,
                        action='store_true',
                        help='Invert y axis')
    parser.add_argument('-sp', '--show_parcel', default=False, action='store_true',
                        help='set to true to plot the parcel path.')
    parser.add_argument('-cape', '--shade_cape', default=False,
                        action='store_true',
                        help='set to true to shade regions of cape.')
    parser.add_argument('-cin', '--shade_cin', default=False, action='store_true',
                        help='set to true to shade regions of cin.')
    parser.add_argument('-d', '--density', default=False, action='store_true',
                        help='Plot a p.d.f. instead of a frequency histogram')
    parser.add_argument('-m', '--mesh', default=False, action='store_true',
                        help=('Set to True to interpolate u and v to '
                              'grid and create wind barbs'))
    parser.add_argument('-uv', '--from_u_and_v', default=False, action='store_true',
                        help='Create SkewTPLot with u and v wind')
    parser.add_argument('-sd', '--from_spd_and_dir', default=False, action='store_true',
                        help='Create SkewTPlot with wind speed and direction')
    parser.add_argument('-px', '--plot_xsection', default=False, action='store_true',
                        help='plots a cross section whose x and y coordinates')
    parser.add_argument('-pxm', '--xsection_map', default=False, action='store_true',
                        help='plots a cross section of 2D data on a geographical map')
    parser.add_argument('-p', '--plot', default=False, action='store_true',
                        help='Makes a time series plot')
    parser.add_argument('-mp', '--multi_panel', default=False,
                        action='store_true',
                        help='Makes a 2 panel timeseries plot')
    parser.add_argument('-qc', '--qc', default=False, action='store_true',
                        help='Create time series plot of embedded quality control values')
    parser.add_argument('-fb', '--fill_between', default=False, action='store_true',
                        help='makes a fill betweem plot based on matplotlib')
    parser.add_argument('-bsd', '--barbs_spd_dir', default=False, action='store_true',
                        help=('Makes time series plot of wind barbs '
                              'using wind speed and dir.'))
    parser.add_argument('-buv', '--barbs_u_v', default=False, action='store_true',
                        help=('Makes time series plot of wind barbs '
                              'using u and v wind components.'))
    parser.add_argument('-pxs', '--xsection_from_1d', default=False,
                        action='store_true',
                        help='Will plot a time-height cross section from 1D dataset')
    parser.add_argument('-ths', '--time_height_scatter',
                        default=False, action='store_true',
                        help='Create a scatter time series plot')
    parser.add_argument('-sbg', '--stacked_bar_graph',
                        default=False, action='store_true',
                        help='Create stacked bar graph histogram')
    parser.add_argument('-psd', '--size_dist', default=False, action='store_true',
                        help='Plots a stairstep plot of size distribution')
    parser.add_argument('-sg', '--stairstep', default=False, action='store_true',
                        help='Plots stairstep plot of a histogram')
    parser.add_argument('-hm', '--heatmap', default=False, action='store_true',
                        help='Plot a heatmap histogram from 2 variables')
    parser.add_argument('-cc', '--create_contour', default=False, action='store_true',
                        help='Extracts, grids, and creates a contour plot')
    parser.add_argument('-cf', '--contourf', default=False, action='store_true',
                        help=('Base function for filled contours if user '
                              'already has data gridded'))
    parser.add_argument('-ct', '--plot_contour', default=False, action='store_true',
                        help=('Base function for contours if user '
                              'already has data gridded'))
    parser.add_argument('-vsd', '--vectors_spd_dir', default=False, action='store_true',
                        help='Extracts, grids, and creates a contour plot.')
    parser.add_argument('-b', '--barbs', default=False, action='store_true',
                        help='Base function for wind barbs.')
    parser.add_argument('-ps', '--plot_station', default=False, action='store_true',
                        help='Extracts, grids, and creates a contour plot')

    # The mutually exclusive but one requried group
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-gp', '--geodisplay', dest='action', action='store_const',
                       const=geodisplay, help='Set to genereate a geographic plot')
    group.add_argument('-skt', '--skewt', dest='action', action='store_const',
                       const=skewt, help='Set to genereate a skew-t plot')
    group.add_argument('-xs', '--xsection', dest='action', action='store_const',
                       const=xsection, help='Set to genereate a XSection plot')
    group.add_argument('-wr', '--wind_rose', dest='action', action='store_const',
                       const=wind_rose, help='Set to genereate a wind rose plot')
    group.add_argument('-ts', '--timeseries', dest='action', action='store_const',
                       const=timeseries, help='Set to genereate a timeseries plot')
    group.add_argument('-c', '--contour', dest='action', action='store_const',
                       const=contour, help='Set to genereate a contour plot')
    group.add_argument('-hs', '--histogram', dest='action', action='store_const',
                       const=histogram, help='Set to genereate a histogram plot')

    args = parser.parse_args()

    # Check if a path but no file name is given. If so use a default name.
    out_path = pathlib.Path(args.out_path)
    if out_path.is_dir():
        args.out_path = str(pathlib.Path(out_path, out_path_default))

    args.action(args)


if __name__ == '__main__':
    main()
